# TODO(kevin): Make this work for all hands.
# TODO(kevin): In the future, we'd like to solve for finger orientation as well.

import copy
import dataclasses
from typing import Mapping, Optional, Sequence, Tuple

import mujoco
import numpy as np
from absl import logging
from dm_control import mjcf
from dm_robotics.geometry import geometry
from dm_robotics.geometry import mujoco_physics

from dexterity import controllers
from dexterity.models.hands import shadow_hand_e_constants as consts
from dexterity.utils import mujoco_utils

# Gain for the linear twist computation, should always be between 0 and 1.
# 0 corresponds to not moving and 1 corresponds to moving to the target in a single
# integration timestep.
_LINEAR_VELOCITY_GAIN = 0.95

# Integration timestep used to convert from joint velocities to joint positions.
_INTEGRATION_TIMESTEP_SEC = 1.0

# Damping factor.
_REGULARIZATION_WEIGHT = 1e-5

# If the norm of the error divided by the magnitude of the joint position update is
# greater than this value, then the solve is ended prematurely. This helps us avoid
# getting stuck in local minima.
_PROGRESS_THRESHOLD = 20.0


@dataclasses.dataclass
class _Solution:
    """Return value of an IK solution."""

    qpos: np.ndarray
    linear_err: float


class IKSolver:
    """Inverse kinematics solver for a dexterous hand."""

    def __init__(
        self,
        model: mjcf.RootElement,
        fingers: Optional[Sequence[consts.Components]] = None,
        prefix: str = "",
    ) -> None:
        """Constructor.

        Args:
            model: The MJCF model root.
            prefix: The prefix assigned to the hand model in case it is attached to
                another entity.
        """
        self._physics = mjcf.Physics.from_mjcf_model(model)
        self._geometry_physics = mujoco_physics.wrap(self._physics)

        if fingers is None:
            self._fingers = consts.FINGERS
        else:
            self._fingers = tuple(fingers)

        # Wrist information.
        wrist_joint_names = [j.name for j in consts.JOINT_GROUP[consts.Components.WR]]
        self._wrist_controllable_joints = []
        for joint_name in wrist_joint_names:
            joint_elem = model.find(
                "joint", mujoco_utils.prefix_identifier(joint_name, prefix)
            )
            self._wrist_controllable_joints.append(joint_elem)
        assert len(self._wrist_controllable_joints) == len(wrist_joint_names)
        self._wrist_num_joints = len(wrist_joint_names)
        self._wirst_joint_bindings = self._physics.bind(self._wrist_controllable_joints)
        self._wrist_nullspace_joint_position_reference = np.zeros(
            self._wrist_num_joints
        )

        # Finger information.
        self._controllable_joints = {}
        self._elements = {}
        for finger in self._fingers:
            fingertip_name = consts.FINGER_FINGERTIP_MAPPING[finger]
            fingertip_site_name = mujoco_utils.prefix_identifier(
                f"{fingertip_name}_site", prefix
            )
            fingertip_site_elem = model.find("site", fingertip_site_name)
            assert fingertip_site_elem is not None
            self._elements[finger] = fingertip_site_elem

            joint_names = [j.name for j in consts.JOINT_GROUP[finger]]
            joint_names += wrist_joint_names
            controllable_joints = []
            for joint_name in joint_names:
                joint_elem = model.find(
                    "joint", mujoco_utils.prefix_identifier(joint_name, prefix)
                )
                controllable_joints.append(joint_elem)
            assert len(controllable_joints) == len(joint_names)
            self._controllable_joints[finger] = controllable_joints

        # Get all the joints of the hand.
        self._all_joints = []
        for joint_name in consts.JOINT_NAMES:
            joint_elem = model.find(
                "joint", mujoco_utils.prefix_identifier(joint_name, prefix)
            )
            assert joint_elem is not None
            self._all_joints.append(joint_elem)

        self._joint_bindings = {}
        self._num_joints = {}
        for finger, controllable_joints in self._controllable_joints.items():
            self._joint_bindings[finger] = self._physics.bind(controllable_joints)
            self._num_joints[finger] = len(controllable_joints)
        self._all_joints_binding = self._physics.bind(self._all_joints)

        self._nullspace_joint_position_reference = np.mean(
            self._all_joints_binding.range, axis=1
        )

        self._create_mapper()

    @property
    def fingers(self) -> Tuple[consts.Components, ...]:
        return self._fingers

    def _create_mapper(self) -> None:
        obj_types = []
        obj_names = []
        for finger in self._fingers:
            obj_types.append(mujoco_utils.get_element_type(self._elements[finger]))
            obj_names.append(self._elements[finger].full_identifier)
        params = controllers.dls.DampedLeastSquaresParameters(
            model=self._physics.model,
            object_types=obj_types,
            object_names=obj_names,
            regularization_weight=_REGULARIZATION_WEIGHT,
        )
        self._mapper = controllers.dls.DampedLeastSquaresMapper(params)

    def solve(
        self,
        target_positions: Mapping[consts.Components, np.ndarray],
        linear_tol: float = 1e-3,
        max_steps: int = 100,
        early_stop: bool = False,
        num_attempts: int = 30,
        stop_on_first_successful_attempt: bool = False,
    ) -> Optional[np.ndarray]:
        nullspace_jnt_qpos_min_err: float = np.inf
        success: bool = False
        sol_qpos: Optional[np.ndarray] = None

        # Each iteration of this loop attempts to solve the IK problem.
        for attempt in range(num_attempts):
            # Randomize the initial joint configuration so that the IK can find
            # different solutions.
            if attempt == 0:
                self._all_joints_binding.qpos[
                    :
                ] = self._nullspace_joint_position_reference
            else:
                self._all_joints_binding.qpos[:] = np.random.uniform(
                    self._all_joints_binding.range[:, 0],
                    self._all_joints_binding.range[:, 1],
                )

            solution = self._solve_ik(
                target_positions,
                linear_tol,
                max_steps,
                early_stop,
            )

            if solution.linear_err <= linear_tol:
                success = True

                nullspace_jnt_qpos_err = float(
                    np.linalg.norm(
                        solution.qpos - self._nullspace_joint_position_reference
                    )
                )
                if nullspace_jnt_qpos_err < nullspace_jnt_qpos_min_err:
                    nullspace_jnt_qpos_min_err = nullspace_jnt_qpos_err
                    sol_qpos = solution.qpos

            if success and stop_on_first_successful_attempt:
                break

        if not success:
            logging.warning(f"{self.__class__.__name__} failed to find a solution.")

        return sol_qpos

    def _solve_ik(
        self,
        target_positions: Mapping[consts.Components, np.ndarray],
        linear_tol: float,
        max_steps: int,
        early_stop: bool,
    ) -> _Solution:
        """Solves for a joint configuration that brings element pose to target pose."""
        cur_frames = {}
        cur_poses = {}
        previous_poses = {}
        for finger, target_position in target_positions.items():
            cur_frame = geometry.PoseStamped(pose=None, frame=self._elements[finger])
            cur_pose = cur_frame.get_world_pose(self._geometry_physics)
            cur_frames[finger] = cur_frame
            cur_poses[finger] = cur_pose
            previous_poses[finger] = copy.copy(cur_pose)

        # Each iteration of this loop attempts to reduce the error between the site's
        # position and the target position.
        for _ in range(max_steps):
            twists = []
            for finger, target_position in target_positions.items():
                twist = _compute_twist(
                    cur_poses[finger],
                    geometry.Pose(position=target_position, quaternion=None),
                    _LINEAR_VELOCITY_GAIN,
                    _INTEGRATION_TIMESTEP_SEC,
                )
                twists.append(twist)

            qdot_sol = self._compute_joint_velocities(twists)

            mujoco.mj_integratePos(
                self._physics.model.ptr,
                self._physics.data.qpos,
                qdot_sol,
                _INTEGRATION_TIMESTEP_SEC,
            )
            self._update_physics_data()

            avg_linear_err: float = 0.0
            close_enough: bool = True
            not_enough_progress: bool = False

            for finger, target_position in target_positions.items():
                # Get the distance between the current pose and the target pose.
                cur_pose = cur_frames[finger].get_world_pose(self._geometry_physics)
                linear_err = float(np.linalg.norm(target_position - cur_pose.position))
                avg_linear_err += linear_err

                # Stop if the pose is close enough to the target pose.
                if linear_err > linear_tol:
                    close_enough = False

                # Stop the solve if not enough progress is being made.
                previous_pose = previous_poses[finger]
                linear_change = np.linalg.norm(
                    cur_pose.position - previous_pose.position
                )
                if linear_err / (linear_change + 1e-10) > _PROGRESS_THRESHOLD:
                    not_enough_progress = True

                previous_poses[finger] = copy.copy(cur_pose)
                cur_poses[finger] = cur_pose

            # Average out the linear error.
            avg_linear_err /= len(target_positions)

            # Break conditions.
            if early_stop and close_enough:
                break
            if not_enough_progress:
                break

        qpos = np.array(self._all_joints_binding.qpos, copy=True)
        return _Solution(qpos=qpos, linear_err=avg_linear_err)

    def _compute_joint_velocities(
        self, cartesian_6d_target: Sequence[np.ndarray]
    ) -> np.ndarray:
        """Maps a Cartesian 6D target velocity to joint velocities."""
        return self._mapper.compute_joint_velocities(
            data=self._physics.data,
            target_velocities=cartesian_6d_target,
            nullspace_bias=None,
        )

    def _update_physics_data(self) -> None:
        """Updates the physics data following the integration of velocities."""
        # Clip joint positions.
        qpos = self._all_joints_binding.qpos
        min_range = self._all_joints_binding.range[:, 0]
        max_range = self._all_joints_binding.range[:, 1]
        qpos = np.clip(qpos, min_range, max_range)
        self._all_joints_binding.qpos[:] = qpos

        # Forward kinematics to update the pose of the tracked site.
        mujoco.mj_normalizeQuat(self._physics.model.ptr, self._physics.data.qpos)
        mujoco.mj_kinematics(self._physics.model.ptr, self._physics.data.ptr)
        mujoco.mj_comPos(self._physics.model.ptr, self._physics.data.ptr)


def _compute_twist(
    init_pose: geometry.Pose,
    final_pose: geometry.Pose,
    linear_velocity_gain: float,
    control_timestep_seconds: float,
) -> np.ndarray:
    """Returns the twist to apply to the element to reach final_pose from init_pose."""
    position_error = final_pose.position - init_pose.position
    linear = linear_velocity_gain * position_error / control_timestep_seconds
    return linear



"""Tasks involving juggling."""

import dataclasses
from typing import Sequence

import numpy as np
from dm_control import composer
from dm_control import mjcf
from dm_control.utils import containers

from dexterity import effector
from dexterity import effectors
from dexterity import task
from dexterity.manipulation import arenas
from dexterity.manipulation import props
from dexterity.manipulation.shared import cameras
from dexterity.manipulation.shared import observations
from dexterity.manipulation.shared import tags
from dexterity.models.hands import adroit_hand
from dexterity.models.hands import dexterous_hand

# The position of the hand relative in the world frame, in meters.
_LEFT_HAND_POS = (-0.1, 0.2, 0.1)
_RIGHT_HAND_POS = (0.1, 0.2, 0.1)

# The orientation of the hand relative to the world frame.
_HAND_QUAT = (0.0, 0.0, 0.707106781186, -0.707106781186)

_BALL_RADIUS = 0.025

# Timestep of the physics simulation.
# OpenAI uses a timestep of 0.002.
_PHYSICS_TIMESTEP: float = 0.02

# Interval between agent actions, in seconds.
# We send a control signal every (_CONTROL_TIMESTEP / _PHYSICS_TIMESTEP) physics steps.
# OpeAI uses a control timestep that is 10x the physics timestep.
_CONTROL_TIMESTEP: float = 0.02  # 50 Hz.

SUITE = containers.TaggedTasks()


class Juggle(task.Task):
    """Juggle a ball with two hands."""

    def __init__(
        self,
        arena: composer.Arena,
        hands: Sequence[dexterous_hand.DexterousHand],
        hand_effectors: Sequence[effector.Effector],
        use_dense_reward: bool,
        control_timestep: float = _CONTROL_TIMESTEP,
        physics_timestep: float = _PHYSICS_TIMESTEP,
    ) -> None:
        """Construct a new `Juggle` task."""

        hand_names = set([hand.name for hand in hands])
        if len(hand_names) != len(hands):
            raise ValueError("Hands must have unique names.")

        super().__init__(arena=arena, hands=hands, hand_effectors=hand_effectors)

        self._use_dense_reward = use_dense_reward

        # Attach the hand to the arena.
        arena.attach_offset(hands[0], position=_LEFT_HAND_POS, quaternion=_HAND_QUAT)
        arena.attach_offset(hands[1], position=_RIGHT_HAND_POS, quaternion=_HAND_QUAT)

        # Add a ball.
        ball = props.JugglingBall(radius=_BALL_RADIUS)
        arena.add_free_entity(ball)
        self._ball = ball

        # Add a closeup camera, used for rendering.
        arena.mjcf_model.worldbody.add(
            "camera", **dataclasses.asdict(cameras.FRONT_CLOSE)
        )

        self.set_timesteps(control_timestep, physics_timestep)

    @property
    def left_hand(self) -> dexterous_hand.DexterousHand:
        return self.hands[0]

    @property
    def right_hand(self) -> dexterous_hand.DexterousHand:
        return self.hands[1]

    @property
    def ball(self):
        return self._ball

    def initialize_episode(
        self, physics: mjcf.Physics, random_state: np.random.RandomState
    ) -> None:
        super().initialize_episode(physics, random_state)

        # Set the initial joint configuration to the midrange of the joint limits.
        midrange = physics.bind(self.left_hand.joints).range.mean(axis=1)
        physics.bind(self.left_hand.joints).qpos[:] = midrange

        midrange = physics.bind(self.right_hand.joints).range.mean(axis=1)
        physics.bind(self.right_hand.joints).qpos[:] = midrange

        # Step the physics to move the fingers out of the way. Typically the pinky
        # collides with the ring finger in this configuration.
        for _ in range(2):
            physics.step()

        # Place the ball on the palm of the left hand.
        palm = self.left_hand.mjcf_model.find("body", "palm")
        palm_position = physics.bind(palm).xpos
        ball_position = palm_position.copy()
        ball_position[1] -= 0.05
        ball_position[2] += 0.05
        self.ball.set_pose(physics, position=ball_position)

    def before_step(
        self,
        physics: mjcf.Physics,
        action: np.ndarray,
        random_state: np.random.RandomState,
    ) -> None:
        super().before_step(physics, action, random_state)

    def after_step(
        self, physics: mjcf.Physics, random_state: np.random.RandomState
    ) -> None:
        super().after_step(physics, random_state)

    def get_reward(self, physics: mjcf.Physics) -> float:
        return 0.0


def juggle_task(
    observation_set: observations.ObservationSet,
    use_dense_reward: bool,
) -> task.Task:
    """Configure and instantiate a `Juggle` task."""
    arena = arenas.Standard()

    left_hand = adroit_hand.AdroitHand(
        observable_options=observations.make_options(
            observation_set.value,
            observations.HAND_OBSERVABLES,
        ),
        name="left_hand",
    )
    left_hand_effector = effectors.HandEffector(left_hand, left_hand.name)

    right_hand = adroit_hand.AdroitHand(
        observable_options=observations.make_options(
            observation_set.value,
            observations.HAND_OBSERVABLES,
        ),
        name="right_hand",
    )
    right_hand_effector = effectors.HandEffector(right_hand, right_hand.name)

    return Juggle(
        arena=arena,
        hands=[left_hand, right_hand],
        hand_effectors=[left_hand_effector, right_hand_effector],
        use_dense_reward=use_dense_reward,
    )


@SUITE.add(tags.STATE, tags.SPARSE)
def state_sparse() -> task.GoalTask:
    """Juggle task with full state observations and sparse reward."""
    return juggle_task(
        observation_set=observations.ObservationSet.STATE_ONLY,
        use_dense_reward=False,
    )



import collections
import inspect
from typing import Optional

from dm_control import composer as _composer

from dexterity import environment as _environment
from dexterity import task as _task
from dexterity.manipulation.tasks import juggle
from dexterity.manipulation.tasks import reach
from dexterity.manipulation.tasks import reorient
from dexterity.utils import mujoco_collisions

# Find all domains imported.
_DOMAINS = {
    name: module
    for name, module in locals().items()
    if inspect.ismodule(module) and hasattr(module, "SUITE")
}


def _get_tasks(tag):
    """Returns a sequence of (domain name, task name) pairs for the given tag."""
    result = []
    for domain_name in sorted(_DOMAINS.keys()):
        domain = _DOMAINS[domain_name]
        if tag is None:
            tasks_in_domain = domain.SUITE
        else:
            tasks_in_domain = domain.SUITE.tagged(tag)
        for task_name in tasks_in_domain.keys():
            result.append((domain_name, task_name))
    return tuple(result)


def _get_tasks_by_domain(tasks):
    """Returns a dict mapping from task name to a tuple of domain names."""
    result = collections.defaultdict(list)

    for domain_name, task_name in tasks:
        result[domain_name].append(task_name)

    return {k: tuple(v) for k, v in result.items()}


# A sequence containing all (domain name, task name) pairs.
ALL_TASKS = _get_tasks(tag=None)

# A sequence containing all `domain_name.task_name` pairs.
ALL_NAMES = [".".join(domain_task) for domain_task in ALL_TASKS]

# A mapping from each domain name to a sequence of its task names.
TASKS_BY_DOMAIN = _get_tasks_by_domain(ALL_TASKS)


def load(
    domain_name: str,
    task_name: str,
    seed: Optional[int] = None,
    strip_singleton_obs_buffer_dim: bool = True,
    time_limit: Optional[float] = None,
) -> _composer.Environment:
    if domain_name not in _DOMAINS:
        raise ValueError(f"Unknown domain: {domain_name}")
    domain = _DOMAINS[domain_name]

    if task_name not in domain.SUITE:
        raise ValueError(f"Unknown task: {task_name}")
    task = domain.SUITE[task_name]()

    # Ensure MuJoCo will not check for collisions between geoms that can never collide.
    mujoco_collisions.exclude_bodies_based_on_contype_conaffinity(
        task.root_entity.mjcf_model
    )

    if isinstance(task, _task.GoalTask):
        env_cls = _environment.GoalEnvironment
    else:
        env_cls = _composer.Environment

    return env_cls(
        task=task,
        time_limit=time_limit or task.time_limit,
        random_state=seed,
        strip_singleton_obs_buffer_dim=strip_singleton_obs_buffer_dim,
    )



from pathlib import Path
from typing import Dict, Tuple

from dexterity import _SRC_ROOT
from dexterity.models.hands.shadow_hand_e_constants import Components

# Path to the shadow hand E series XML file.
ADROIT_HAND_E_XML: Path = (
    _SRC_ROOT
    / "models"
    / "vendor"
    / "adroit"
    / "adroit_hand_description"
    / "mjcf"
    / "adroit_hand.xml"
)

FINGERTIP_SITE_NAMES: Tuple[str, ...] = (
    "S_fftip",
    "S_mftip",
    "S_rftip",
    "S_lftip",
    "S_thtip",
)

# Mapping from finger `Components` to its associated geoms.
FINGER_GEOM_MAPPING: Dict[Components, Tuple[str, ...]] = {
    Components.FF: ("V_ffproximal", "V_ffmiddle", "V_ffdistal"),
    Components.MF: ("V_mfproximal", "V_mfmiddle", "V_mfdistal"),
    Components.RF: ("V_rfproximal", "V_rfmiddle", "V_rfdistal"),
    Components.LF: ("V_lfproximal", "V_lfmiddle", "V_lfdistal"),
    Components.TH: ("V_thproximal", "V_thmiddle", "V_thdistal"),
}



import numpy as np
from dm_control import mjcf
from dm_control.composer.variation import rotations
from dm_control.entities.props import primitive
from dm_env import specs
from dm_robotics.transformations import transformations as tr

from dexterity import goal


class PropOrientation(goal.GoalGenerator):
    def __init__(
        self,
        prop: primitive.Primitive,
        name: str = "prop_orientation_goal_generator",
    ) -> None:
        super().__init__()

        self._name = name
        self._prop = prop
        self._sampler = rotations.UniformQuaternion()

    def goal_spec(self) -> specs.Array:
        return specs.Array(shape=(4,), dtype=np.float64, name=self._name)

    def initialize_episode(
        self, physics: mjcf.Physics, random_state: np.random.RandomState
    ) -> None:
        del physics, random_state  # Unused.

    def current_state(self, physics: mjcf.Physics) -> np.ndarray:
        return np.array(physics.bind(self._prop.orientation).sensordata)

    def next_goal(
        self, physics: mjcf.Physics, random_state: np.random.RandomState
    ) -> np.ndarray:
        del physics  # Unused.
        return self._sampler(random_state)

    def relative_goal(
        self, goal_state: np.ndarray, current_state: np.ndarray
    ) -> np.ndarray:
        return tr.quat_diff_active(source_quat=current_state, target_quat=goal_state)

    def goal_distance(
        self, goal_state: np.ndarray, current_state: np.ndarray
    ) -> np.ndarray:
        err_quat = self.relative_goal(goal_state, current_state)
        err_axisangle = tr.quat_to_axisangle(err_quat)
        return np.linalg.norm(err_axisangle, keepdims=True)  # type: ignore

    @property
    def name(self) -> str:
        return self._name



"""Utility functions for dealing with MuJoCo actuators."""

from typing import Tuple

from dexterity import hints


def add_position_actuator(
    elem: hints.MjcfElement,
    elem_type: str,
    qposrange: Tuple[float, float],
    ctrlrange: Tuple[float, float] = (-1.0, 1.0),
    kp: float = 1.0,
    **kwargs,
) -> hints.MjcfElement:
    """Adds a scaled position actuator that is bound to the specified element.

    This is equivalent to MuJoCo's built-in `<position>` actuator where an affine
    transformation is pre-applied to the control signal, such that the minimum control
    value corresponds to the minimum desired position, and the maximum control value
    corresponds to the maximum desired position.

    Args:
        elem: The joint or tendon element that is to be actuated.
        elem_type: The type of the element. Must be either `"joint"` or `"tendon"`.
        qposrange: A sequence of two numbers specifying the allowed range of target
            position.
        ctrlrange: A sequence of two numbers specifying the allowed range of this
            actuator's control signal.
        kp: The gain parameter of this position actuator.
        **kwargs:

    Returns:
        The actuator element.
    """
    kwargs[elem_type] = elem
    slope = (qposrange[1] - qposrange[0]) / (ctrlrange[1] - ctrlrange[0])
    g0 = kp * slope
    b0 = kp * (qposrange[0] - slope * ctrlrange[0])
    b1 = -kp
    b2 = 0
    return elem.root.actuator.add(
        "general",
        biastype="affine",
        gainprm=[g0],
        biasprm=[b0, b1, b2],
        ctrllimited=True,
        ctrlrange=ctrlrange,
        **kwargs,
    )



import abc
from typing import List

import numpy as np
from dm_control import composer
from dm_control import mjcf
from dm_control.composer.observation import observable
from dm_robotics.transformations import transformations as tr

from dexterity.hints import MjcfElement
from dexterity.utils import mujoco_utils


class DexterousHand(abc.ABC, composer.Entity):
    """Base composer class for a dexterous multi-fingered hand."""

    def _build_observables(self) -> composer.Observables:
        return DexterousHandObservables(self)

    @abc.abstractmethod
    def _build(self) -> None:
        """Entity initialization method to be overridden by subclasses."""

    @property
    @abc.abstractmethod
    def mjcf_model(self) -> mjcf.RootElement:
        """Returns the `mjcf.RootElement` object corresponding to the hand."""

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """The name of the hand."""

    @property
    @abc.abstractmethod
    def actuators(self) -> List[MjcfElement]:
        """List of actuator elements belonging to the hand."""

    @property
    @abc.abstractmethod
    def joints(self) -> List[MjcfElement]:
        """List of joint elements belonging to the hand."""

    @property
    @abc.abstractmethod
    def fingertip_sites(self) -> List[MjcfElement]:
        """List of fingertip site elements belonging to the hand."""

    @property
    @abc.abstractmethod
    def joint_torque_sensors(self) -> List[MjcfElement]:
        """List of joint torque sensor elements belonging to the hand."""

    @abc.abstractmethod
    def control_to_joint_positions(self, control: np.ndarray) -> np.ndarray:
        """Maps a control command to a joint position command.

        This method is necessary for underactuated hands.
        """

    @abc.abstractmethod
    def joint_positions_to_control(self, qpos: np.ndarray) -> np.ndarray:
        """Maps a joint position command to a control command.

        This method is necessary for underactuated hands.
        """

    # Note: This method is abstract because a hand might have to perform extra logic
    # specific to its actuation after setting the joint angles.
    @abc.abstractmethod
    def set_joint_angles(self, physics: mjcf.Physics, joint_angles: np.ndarray) -> None:
        """Sets the joints of the hand to a given configuration."""

    # Note: This method is abstract because a hand might have to perform extra logic
    # pertaining to its underactuation after randomly sampling joint angles.
    @abc.abstractmethod
    def sample_joint_angles(
        self, physics: mjcf.Physics, random_state: np.random.RandomState
    ) -> np.ndarray:
        """Samples a random joint configuration for the hand."""


class DexterousHandObservables(composer.Observables):
    """Observables for a fingered hand."""

    _entity: DexterousHand

    @composer.observable
    def joint_positions(self) -> observable.MJCFFeature:
        return observable.MJCFFeature(kind="qpos", mjcf_element=self._entity.joints)

    @composer.observable
    def joint_positions_sin_cos(self) -> observable.MJCFFeature:
        def _get_joint_pos_sin_cos(physics: mjcf.Physics) -> np.ndarray:
            qpos = physics.bind(self._entity.joints).qpos
            qpos_sin = np.sin(qpos)
            qpos_cos = np.cos(qpos)
            return np.concatenate([qpos_sin, qpos_cos])

        return observable.Generic(raw_observation_callable=_get_joint_pos_sin_cos)

    @composer.observable
    def joint_velocities(self) -> observable.MJCFFeature:
        return observable.MJCFFeature(kind="qvel", mjcf_element=self._entity.joints)

    @composer.observable
    def joint_torques(self) -> observable.Generic:
        def _get_joint_torques(physics: mjcf.Physics) -> np.ndarray:
            torques = physics.bind(self._entity.joint_torque_sensors).sensordata
            joint_axes = physics.bind(self._entity.joints).axis
            joint_torques = np.einsum("ij,ij->i", torques.reshape(-1, 3), joint_axes)
            return joint_torques

        return observable.Generic(raw_observation_callable=_get_joint_torques)

    @composer.observable
    def fingertip_positions(self) -> observable.Generic:
        def _get_fingertip_positions(physics: mjcf.Physics) -> np.ndarray:
            return physics.bind(self._entity.fingertip_sites).xpos.ravel()

        return observable.Generic(raw_observation_callable=_get_fingertip_positions)

    @composer.observable
    def fingertip_orientations(self) -> observable.Generic:
        def _get_fingertip_orientations(physics: mjcf.Physics) -> np.ndarray:
            xmats = physics.bind(self._entity.fingertip_sites).xmat.reshape(-1, 3, 3)
            quats = [tr.mat_to_quat(xmat) for xmat in xmats]
            return np.concatenate(quats)

        return observable.Generic(raw_observation_callable=_get_fingertip_orientations)

    @composer.observable
    def fingertip_linear_velocities(self) -> observable.Generic:
        def _get_fingertip_linear_velocities(physics: mjcf.Physics) -> np.ndarray:
            linear_velocities = []
            for fingertip_site in self._entity.fingertip_sites:
                lin_vel = mujoco_utils.get_site_velocity(physics, fingertip_site)
                linear_velocities.append(lin_vel[:3])
            return np.concatenate(linear_velocities)

        return observable.Generic(
            raw_observation_callable=_get_fingertip_linear_velocities
        )

    @composer.observable
    def fingertip_angular_velocities(self) -> observable.Generic:
        def _get_fingertip_angular_velocities(physics: mjcf.Physics) -> np.ndarray:
            angular_velocities = []
            for fingertip_site in self._entity.fingertip_sites:
                ang_vel = mujoco_utils.get_site_velocity(physics, fingertip_site)
                angular_velocities.append(ang_vel[3:])
            return np.concatenate(angular_velocities)

        return observable.Generic(
            raw_observation_callable=_get_fingertip_angular_velocities
        )



from typing import Tuple

from dm_control import composer
from dm_control import mjcf


class TargetSphere(composer.Entity):
    """A non-colliding spherical site that can be used as a fingertip target."""

    def _build(
        self,
        radius: float,
        rgba: Tuple[float, float, float, float],
        name: str = "target",
    ) -> None:
        self._mjcf_root = mjcf.RootElement(model=name)

        self._site = self._mjcf_root.worldbody.add(
            "site",
            type="sphere",
            size=[radius],
            rgba=rgba,
            group=None,  # Make the site visible.
        )

    @property
    def mjcf_model(self):
        return self._mjcf_root

    @property
    def site(self):
        return self._site



"""Tasks involving in-hand object re-orientation."""

import collections
import dataclasses
from typing import Dict, Optional

import numpy as np
from dm_control import composer
from dm_control import mjcf
from dm_control.composer import initializers
from dm_control.composer.variation import distributions
from dm_control.composer.variation import rotations
from dm_control.entities.props import primitive
from dm_control.utils import containers
from dm_control.utils import rewards as reward_utils
from dm_robotics.transformations import transformations as tr

from dexterity import effector
from dexterity import effectors
from dexterity import goal
from dexterity import task
from dexterity.manipulation import arenas
from dexterity.manipulation import props
from dexterity.manipulation.goals import prop_orientation
from dexterity.manipulation.shared import cameras
from dexterity.manipulation.shared import constants
from dexterity.manipulation.shared import observations
from dexterity.manipulation.shared import rewards
from dexterity.manipulation.shared import tags
from dexterity.manipulation.shared import workspaces
from dexterity.models.hands import dexterous_hand
from dexterity.models.hands import shadow_hand_e
from dexterity.utils import mujoco_collisions


@dataclasses.dataclass(frozen=True)
class Workspace:
    prop_bbox: workspaces.BoundingBox


# The position of the hand relative in the world frame, in meters.
_HAND_POS = (0, 0.2, 0.1)
# The orientation of the hand relative to the world frame.
_HAND_QUAT = tr.axisangle_to_quat(
    np.pi * np.array([0, np.sqrt(2) / 2, -np.sqrt(2) / 2])
)

# Alpha value of the visual goal hint.
_HINT_ALPHA = 0.4
# Position of the hint in the world frame, in meters.
_HINT_POS = (0.12, 0.0, 0.15)

# Size of the prop, in meters.
_PROP_SIZE = 0.02

# Fudge factor for taking the inverse of the orientation error, in radians.
_ORIENTATION_EPS = 0.1
# Threshold for successful orientation, in radians.
_ORIENTATION_THRESHOLD = 0.1
# Reward shaping coefficients.
_ORIENTATION_WEIGHT = 1.0
_SUCCESS_BONUS_WEIGHT = 800.0
_ACTION_SMOOTHING_WEIGHT = -0.1  # NOTE(kevin): negative sign.

# Timestep of the physics simulation.
_PHYSICS_TIMESTEP: float = 0.005

# Interval between agent actions, in seconds.
_CONTROL_TIMESTEP: float = 0.025

# The maximum number of consecutive solves until the task is terminated.
_SUCCESSED_NEEDED: int = 1

# The maximum allowed time for reaching the current target, in seconds.
_MAX_STEPS_SINGLE_SOLVE: int = 300
_MAX_TIME_SINGLE_SOLVE: float = _MAX_STEPS_SINGLE_SOLVE * _CONTROL_TIMESTEP

_STEPS_BEFORE_MOVING_TARGET: int = 5

_BBOX_SIZE = 0.05
_WORKSPACE = Workspace(
    prop_bbox=workspaces.BoundingBox(
        lower=(-_BBOX_SIZE / 2, -0.13 - _BBOX_SIZE / 2, 0.16),
        upper=(+_BBOX_SIZE / 2, -0.13 + _BBOX_SIZE / 2, 0.16),
    ),
)

# Observable settings.
_FREEPROP_OBSERVABLES = observations.ObservableNames(
    prop_pose=("position", "orientation", "linear_velocity", "angular_velocity"),
)
_TARGETPROP_OBSERVABLES = observations.ObservableNames(
    prop_pose=("orientation",),
)

SUITE = containers.TaggedTasks()


class ReOrient(task.GoalTask):
    """Manipulate an object to a goal orientation."""

    def __init__(
        self,
        arena: composer.Arena,
        hand: dexterous_hand.DexterousHand,
        hand_effector: effector.Effector,
        goal_generator: goal.GoalGenerator,
        prop: primitive.Primitive,
        hint_prop: primitive.Primitive,
        workspace: Workspace = _WORKSPACE,
        fall_termination: bool = True,
        success_threshold: float = _ORIENTATION_THRESHOLD,
        successes_needed: int = _SUCCESSED_NEEDED,
        steps_before_changing_goal: int = _STEPS_BEFORE_MOVING_TARGET,
        max_time_per_goal: Optional[float] = _MAX_TIME_SINGLE_SOLVE,
        control_timestep: float = _CONTROL_TIMESTEP,
        physics_timestep: float = _PHYSICS_TIMESTEP,
    ) -> None:
        """Construct a new `ReOrient` task."""

        super().__init__(
            arena=arena,
            hands=[hand],
            hand_effectors=[hand_effector],
            goal_generator=goal_generator,
            success_threshold=success_threshold,
            successes_needed=successes_needed,
            steps_before_changing_goal=steps_before_changing_goal,
            max_time_per_goal=max_time_per_goal,
        )

        self._fall_termination = fall_termination

        # Attach the hand to the arena.
        arena.attach_offset(hand, position=_HAND_POS, quaternion=_HAND_QUAT)

        # Add prop to the arena.
        arena.add_free_entity(prop)
        self._prop = prop

        # Translucent, contactless prop with no observables. This is used to provide a
        # visual hint of the goal state.
        _hintify(hint_prop, _HINT_ALPHA)
        arena.attach_offset(hint_prop, position=_HINT_POS)
        self._hint_prop = hint_prop

        self._prop_placer = initializers.PropPlacer(
            props=[self._prop],
            position=distributions.Uniform(
                workspace.prop_bbox.lower,
                workspace.prop_bbox.upper,
            ),
            quaternion=rotations.UniformQuaternion(),
            settle_physics=False,
        )

        # Add a closeup camera, used for rendering.
        arena.mjcf_model.worldbody.add(
            "camera", **dataclasses.asdict(cameras.FRONT_CLOSE)
        )

        # Visual debugging.
        workspaces.add_bbox_site(
            body=self.root_entity.mjcf_model.worldbody,
            lower=workspace.prop_bbox.lower,
            upper=workspace.prop_bbox.upper,
            rgba=constants.GREEN,
            name="prop_spawn_area",
            visible=False,
        )

        self.set_timesteps(control_timestep, physics_timestep)

    @property
    def hand(self) -> dexterous_hand.DexterousHand:
        return self.hands[0]

    @property
    def hand_effector(self) -> effector.Effector:
        return self.hand_effectors[0]

    @property
    def root_entity(self) -> composer.Entity:
        return self._arena

    def initialize_episode(
        self, physics: mjcf.Physics, random_state: np.random.RandomState
    ) -> None:
        super().initialize_episode(physics, random_state)

        self._hint_prop.set_pose(physics=physics, quaternion=self._goal)
        self._prop_placer(physics=physics, random_state=random_state)

    def before_step(
        self,
        physics: mjcf.Physics,
        action: np.ndarray,
        random_state: np.random.RandomState,
    ) -> None:
        super().before_step(physics, action, random_state)

        if self._goal_changed:
            self._hint_prop.set_pose(physics=physics, quaternion=self._goal)

    def after_step(
        self, physics: mjcf.Physics, random_state: np.random.RandomState
    ) -> None:
        super().after_step(physics, random_state)

        self._failure_termination = False
        if self._fall_termination:
            if self._is_prop_fallen(physics):
                self._failure_termination = True

    def should_terminate_episode(self, physics: mjcf.Physics) -> bool:
        should_terminate = super().should_terminate_episode(physics)
        return should_terminate or self._failure_termination

    def get_reward(self, physics: mjcf.Physics) -> float:
        shaped_reward = _get_shaped_reorientation_reward(
            physics,
            self._goal_distance,
        )
        return rewards.weighted_average(shaped_reward)

    def get_discount(self, physics: mjcf.Physics) -> float:
        if self._failure_termination:
            return 1.0
        return super().get_discount(physics)

    # Helper methods.

    def _is_prop_fallen(self, physics: mjcf.Physics) -> bool:
        """Returns True if the prop has fallen from the hand."""
        return mujoco_collisions.has_collision(
            physics=physics,
            collision_geom_prefix_1=[f"{self._prop.name}/"],
            collision_geom_prefix_2=[self._arena.ground.full_identifier],
        )


def _get_shaped_reorientation_reward(
    physics: mjcf.Physics, goal_distance: np.ndarray
) -> Dict[str, rewards.Reward]:
    """Returns a tuple of shaping reward components, as defined in [1].

    The reward is a weighted sum of the following components:
        - orientation reward: The inverse of the absolute value of the angular error
            between the prop's current orientation and the goal orientation.
        - success reward: 1.0 if the the angular error is within a tolerance and 0.0
            otherwise.
        - action smoothness reward: The negative of the squared L2 norm of the control
            action.

    References:
        [1]: A System for General In-Hand Object Re-Orientation,
        https://arxiv.org/abs/2111.03043
    """
    shaped_reward = collections.OrderedDict()

    # Orientation component.
    distance = float(goal_distance[0])
    orientation_reward = 1.0 / (distance + _ORIENTATION_EPS)
    shaped_reward["orientation"] = rewards.Reward(
        value=orientation_reward, weight=_ORIENTATION_WEIGHT
    )

    # Success bonus component.
    success_bonus_reward = reward_utils.tolerance(
        x=distance,
        bounds=(0, _ORIENTATION_THRESHOLD),
        margin=0.0,
    )
    assert isinstance(success_bonus_reward, float)
    shaped_reward["success_bonus"] = rewards.Reward(
        value=success_bonus_reward,
        weight=_SUCCESS_BONUS_WEIGHT,
    )

    # Action smoothing component.
    action_smoothing_reward = np.linalg.norm(physics.data.ctrl) ** 2
    assert isinstance(action_smoothing_reward, float)
    shaped_reward["action_smoothing"] = rewards.Reward(
        value=action_smoothing_reward,
        weight=_ACTION_SMOOTHING_WEIGHT,
    )

    return shaped_reward


def _replace_alpha(rgba: np.ndarray, alpha: float = 0.3) -> np.ndarray:
    """Replaces the alpha value of a color tuple."""
    new_rgba = rgba.copy()
    new_rgba[3] = alpha
    return new_rgba


def _hintify(entity: composer.Entity, alpha: Optional[float] = None) -> None:
    """Modifies an entity for use as a visual hint.

    Specifically, contacts are disabled for all geoms within the entity, and its bodies
    are converted to mocap bodies which are viewed as fixed from the perspective of the
    dynamics. Additionally, the geom alpha values can be overriden to render the geoms
    as translucent.
    """
    for subentity in entity.iter_entities():
        if (
            alpha is not None
            and subentity.mjcf_model.default.geom is not None
            and subentity.mjcf_model.default.geom.rgba is not None
        ):
            subentity.mjcf_model.default.geom.rgba = _replace_alpha(
                subentity.mjcf_model.default.geom.rgba, alpha=alpha
            )
        for body in subentity.mjcf_model.find_all("body"):
            body.mocap = "true"
        for geom in subentity.mjcf_model.find_all("geom"):
            if alpha is not None and geom.rgba is not None:
                geom.rgba = _replace_alpha(geom.rgba, alpha=alpha)
            # This deals with textures.
            if alpha is not None and geom.material is not None:
                material = subentity.mjcf_model.find("material", geom.material)
                material.rgba = _replace_alpha(material.rgba, alpha=alpha)
            geom.contype = 0
            geom.conaffinity = 0


def reorient_task(
    observation_set: observations.ObservationSet,
) -> composer.Task:
    """Configure and instantiate a `ReOrient` task."""
    arena = arenas.Standard()

    hand = shadow_hand_e.ShadowHandSeriesE(
        observable_options=observations.make_options(
            observation_set.value,
            observations.HAND_OBSERVABLES,
        ),
    )

    hand_effector = effectors.HandEffector(hand=hand, hand_name=hand.name)

    prop_obs_options = observations.make_options(
        observation_set.value, _FREEPROP_OBSERVABLES
    )
    prop = props.OpenAICube(
        size=_PROP_SIZE, observable_options=prop_obs_options, name="prop"
    )

    target_prop_obs_options = observations.make_options(
        observation_set.value, _TARGETPROP_OBSERVABLES
    )
    hint_prop = props.OpenAICube(
        size=_PROP_SIZE,
        observable_options=target_prop_obs_options,
        name="target_prop",
    )

    goal_generator = prop_orientation.PropOrientation(prop=prop)

    return ReOrient(
        arena=arena,
        hand=hand,
        hand_effector=hand_effector,
        prop=prop,
        hint_prop=hint_prop,
        goal_generator=goal_generator,
    )


@SUITE.add(tags.STATE)
def state_dense() -> composer.Task:
    return reorient_task(
        observation_set=observations.ObservationSet.STATE_ONLY,
    )



import numpy as np
from dm_control import mjcf
from dm_env import specs

from dexterity import effector
from dexterity.effectors import mujoco_actuation
from dexterity.models.hands import dexterous_hand


class HandEffector(effector.Effector):
    def __init__(
        self,
        hand: dexterous_hand.DexterousHand,
        hand_name: str,
    ) -> None:
        self._hand = hand
        self._effector_prefix = f"{hand_name}_joint"

        self._mujoco_effector = mujoco_actuation.MujocoEffector(
            actuators=self._hand.actuators, prefix=self._effector_prefix
        )

    def after_compile(self, mjcf_model: mjcf.RootElement) -> None:
        del mjcf_model  # Unused.

    def initialize_episode(self, physics, random_state) -> None:
        del physics, random_state  # Unused.

    def action_spec(self, physics: mjcf.Physics) -> specs.BoundedArray:
        return self._mujoco_effector.action_spec(physics)

    def set_control(self, physics: mjcf.Physics, command: np.ndarray) -> None:
        self._mujoco_effector.set_control(physics, command)

    @property
    def prefix(self) -> str:
        return self._mujoco_effector.prefix



"""Tools for adding custom cameras to the arena."""

import collections
import dataclasses
from typing import Tuple

from dm_control import composer
from dm_control.composer.observation import observable

from dexterity.manipulation.shared import observations


@dataclasses.dataclass(frozen=True)
class CameraConfig:
    name: str
    pos: Tuple[float, float, float]
    xyaxes: Tuple[float, float, float, float, float, float]


# Custom cameras that can be added to the arena.

FRONT_CLOSE = CameraConfig(
    name="front_close",
    pos=(0.0, -0.5, 0.5),
    xyaxes=(1.0, 0.0, 0.0, 0.0, 0.7, 0.75),
)

LEFT_CLOSE = CameraConfig(
    name="left_close",
    pos=(-0.6, 0.0, 0.5),
    xyaxes=(0.0, -1.0, 0.0, 0.7, 0.0, 0.75),
)

RIGHT_CLOSE = CameraConfig(
    name="right_close",
    pos=(0.6, 0.0, 0.5),
    xyaxes=(0.0, 1.0, 0.0, -0.7, 0.0, 0.75),
)

FRONT_FAR = CameraConfig(
    name="front_far",
    pos=(0.0, -1.0, 0.7),
    xyaxes=(1.0, 0.0, 0.0, 0.0, 0.7, 0.75),
)

TOP_DOWN = CameraConfig(
    name="top_down",
    pos=(0.0, 0.0, 2.5),
    xyaxes=(1.0, 0.0, 0.0, 0.0, 1.0, 0.0),
)


def add_camera_observables(
    entity: composer.Entity,
    obs_settings: observations.ObservationSettings,
    *camera_configs: CameraConfig,
) -> collections.OrderedDict:
    obs_dict = collections.OrderedDict()
    for config in camera_configs:
        camera = entity.mjcf_model.worldbody.add("camera", **dataclasses.asdict(config))
        obs = observable.MJCFCamera(camera)
        obs.configure(**dataclasses.asdict(obs_settings.camera))
        obs_dict[config.name] = obs
    return obs_dict



from dexterity.controllers import dls
from dexterity.controllers.mapper import CartesianVelocitytoJointVelocityMapper
from dexterity.controllers.mapper import Parameters

__all__ = [
    "CartesianVelocitytoJointVelocityMapper",
    "Parameters",
    "dls",
]



from dexterity.controllers.dls.dls import DampedLeastSquaresMapper
from dexterity.controllers.dls.dls import DampedLeastSquaresParameters

__all__ = [
    "DampedLeastSquaresMapper",
    "DampedLeastSquaresParameters",
]



import dm_env
import numpy as np
from dm_env import specs

from dexterity.manipulation.wrappers import base


# Adapted from https://github.com/deepmind/dm_control/blob/main/dm_control/suite/wrappers/action_noise.py
class ActionNoise(base.Wrapper):
    """A wrapper that adds Gaussian noise to the actions."""

    def __init__(self, environment: dm_env.Environment, scale: float):
        super().__init__(environment)

        action_spec = environment.action_spec()
        if not isinstance(action_spec, specs.BoundedArray):
            raise ValueError("Only `BoundedArray` action specs are supported.")

        self._minimum = action_spec.minimum
        self._maximum = action_spec.maximum
        self._noise_stddev = scale * (action_spec.maximum - action_spec.minimum)

    def step(self, action: np.ndarray) -> dm_env.TimeStep:
        noisy_action = action + self.random_state.normal(scale=self._noise_stddev)
        np.clip(noisy_action, self._minimum, self._maximum, out=noisy_action)
        return self.environment.step(noisy_action)



import dataclasses
from typing import Optional, Sequence

import numpy as np

from dexterity import hints
from dexterity.controllers import mapper
from dexterity.utils import mujoco_utils


@dataclasses.dataclass
class DampedLeastSquaresParameters(mapper.Parameters):

    regularization_weight: float
    """Damping factor used to regularize the pseudoinverse."""

    def validate_parameters(self) -> None:
        if self.regularization_weight < 0:
            raise ValueError(
                "`regularization_weight` must be non-negative, but was "
                f"{self.regularization_weight}."
            )

        super().validate_parameters()


@dataclasses.dataclass
class DampedLeastSquaresMapper(mapper.CartesianVelocitytoJointVelocityMapper):
    """A `CartesianVelocitytoJointVelocityMapper` that uses damped least-squares to
    solve for the joint velocities.

    Specifically, it maps the desired Cartesian velocity V to desired joint velocities
    v using the relationship:

    `J^T v = [J^T @ J + λI] @ V`

    where @ denotes the matrix product, J is the end-effector Jacobian, λ is a damping
    factor and I is the identity matrix.
    """

    params: DampedLeastSquaresParameters

    def __post_init__(self) -> None:
        self.params.validate_parameters()

    def compute_joint_velocities(
        self,
        data: hints.MjData,
        target_velocities: Sequence[np.ndarray],
        nullspace_bias: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        del nullspace_bias

        jacobians = []
        for obj_type, obj_name in zip(
            self.params.object_types, self.params.object_names
        ):
            jacobian = mujoco_utils.compute_object_6d_jacobian(
                model=self.params.model,
                data=data,
                object_type=obj_type,
                object_id=self.params.model.name2id(obj_name, obj_type),
            )

            # TODO(kevin): This logic is specific to the Shadow Hand and should be
            # abstracted out for other hands.
            # Underactuation correction.
            avg = 0.5 * (jacobian[:, 4] + jacobian[:, 5])
            jacobian[:, 4] = jacobian[:, 5] = avg
            avg = 0.5 * (jacobian[:, 8] + jacobian[:, 9])
            jacobian[:, 8] = jacobian[:, 9] = avg
            avg = 0.5 * (jacobian[:, 12] + jacobian[:, 13])
            jacobian[:, 12] = jacobian[:, 13] = avg
            avg = 0.5 * (jacobian[:, 17] + jacobian[:, 18])
            jacobian[:, 17] = jacobian[:, 18] = avg

            jacobians.append(jacobian[:3])
        jacobian = np.concatenate(jacobians, axis=0)
        twist = np.concatenate(target_velocities, axis=0)

        # Solve!
        hess_approx = jacobian.T @ jacobian
        hess_approx += np.eye(hess_approx.shape[0]) * self.params.regularization_weight
        return np.linalg.solve(hess_approx, jacobian.T @ twist)



"""Tests for reorient.py."""

import numpy as np
from absl.testing import absltest

from dexterity import environment
from dexterity.manipulation.shared import observations
from dexterity.manipulation.tasks import reorient
from dexterity.manipulation.tasks import reorient_task


class ReOrientTaskTest(absltest.TestCase):
    def test_dense_reward(self) -> None:
        task = reorient_task(observations.ObservationSet.STATE_ONLY)

        random_state = np.random.RandomState(12345)
        env = environment.GoalEnvironment(task, random_state=random_state)
        action_spec = env.action_spec()
        timestep = env.reset()
        self.assertIsNone(timestep.reward)

        # Manually set the prop's orientation to be equal to the goal orientation.
        env.task._prop.set_pose(physics=env.physics, quaternion=env.task._goal)

        # Artificially set the ctrl to test its reward component.
        rand_ctrl = np.random.uniform(
            low=action_spec.minimum,
            high=action_spec.maximum,
            size=action_spec.shape,
        )
        env.task.hand_effector.set_control(env.physics, rand_ctrl)

        # Compute shaped reward.
        goal_distance = env.task.goal_generator.goal_distance(
            env.task._goal,
            env.task.goal_generator.current_state(env.physics),
        )
        shaped_reward = reorient._get_shaped_reorientation_reward(
            physics=env.physics,
            goal_distance=goal_distance,
        )

        # Check individual reward components.
        np.testing.assert_equal(
            shaped_reward["orientation"].value, 1 / reorient._ORIENTATION_THRESHOLD
        )
        np.testing.assert_equal(shaped_reward["success_bonus"].value, 1.0)
        np.testing.assert_equal(
            shaped_reward["action_smoothing"].value, np.linalg.norm(rand_ctrl) ** 2
        )


if __name__ == "__main__":
    absltest.main()



"""Shadow hand constants."""

import enum
from math import radians as rad
from pathlib import Path
from typing import Dict, List, Tuple

import numpy as np

from dexterity import _SRC_ROOT

# Path to the shadow hand E series XML file.
SHADOW_HAND_E_XML: Path = (
    _SRC_ROOT
    / "models"
    / "vendor"
    / "shadow_robot"
    / "shadow_hand_description"
    / "mjcf"
    / "shadow_hand_series_e.xml"
)


class Components(enum.Enum):
    """The actuated components of the hand: wrist and fingers."""

    WR = "wrist"
    FF = "first_finger"
    MF = "middle_finger"
    RF = "ring_finger"
    LF = "little_finger"
    TH = "thumb"


FINGERS: Tuple[Components, ...] = (
    Components.FF,
    Components.MF,
    Components.RF,
    Components.LF,
    Components.TH,
)

# ====================== #
# Joint constants
# ====================== #


class Joints(enum.Enum):
    """Joints of the Shadow Hand.

    There are a total of 24 joints:
        * 2 joints for the wrist
        * 4 joints for the first, middle and ring fingers (4 * 3 = 12)
        * 5 joints for the little finger and thumb (5 * 2 = 10)

    The joint numbering is increasing from fingertip to palm, i.e. FFJ0 is the first
    knuckle of the first finger, FFJ3 is the last knuckle of the first finger, etc.

    The first two joints (*FJ0, *FJ1) of the main fingers (exlcuding the thumb) are
    coupled, which means 1 actuator controls both of them via a tendon. This means there
    is 1 less DoF for each of these fingers, which means the total number of degrees of
    freedom for the hand is: 24 - 4 = 20.
    """

    # Wrist: 2 joints, 2 degrees of freedom.
    WRJ1 = enum.auto()
    WRJ0 = enum.auto()
    # First finger: 4 joints, 3 degrees of freedom.
    FFJ3 = enum.auto()
    FFJ2 = enum.auto()
    FFJ1 = enum.auto()  # c 4
    FFJ0 = enum.auto()  # c 5
    # Middle finger: 4 joints, 3 degrees of freedom.
    MFJ3 = enum.auto()
    MFJ2 = enum.auto()
    MFJ1 = enum.auto()  # c 8
    MFJ0 = enum.auto()  # c 9
    # Ring finger: 4 joints, 3 degrees of freedom.
    RFJ3 = enum.auto()
    RFJ2 = enum.auto()
    RFJ1 = enum.auto()  # c 12
    RFJ0 = enum.auto()  # c 13
    # Little finger: 5 joints, 4 degrees of freedom.
    LFJ4 = enum.auto()
    LFJ3 = enum.auto()
    LFJ2 = enum.auto()
    LFJ1 = enum.auto()  # c 17
    LFJ0 = enum.auto()  # c 18
    # Thumb: 5 joints, 5 degrees of freedom.
    THJ4 = enum.auto()
    THJ3 = enum.auto()
    THJ2 = enum.auto()
    THJ1 = enum.auto()
    THJ0 = enum.auto()


# The total number of joints.
NUM_JOINTS: int = len(Joints)

# A list of joint names, as strings.
JOINT_NAMES: List[str] = [j.name for j in Joints]

# A mapping from `Components` to the list of `Joints` that belong to it.
JOINT_GROUP: Dict[Components, Tuple[Joints, ...]] = {
    # Wrist has 2 joints.
    Components.WR: (Joints.WRJ1, Joints.WRJ0),
    # First, middle and ring fingers have 4 joints.
    Components.FF: (Joints.FFJ3, Joints.FFJ2, Joints.FFJ1, Joints.FFJ0),
    Components.MF: (Joints.MFJ3, Joints.MFJ2, Joints.MFJ1, Joints.MFJ0),
    Components.RF: (Joints.RFJ3, Joints.RFJ2, Joints.RFJ1, Joints.RFJ0),
    # Little finger has 5 joints.
    Components.LF: (
        Joints.LFJ4,
        Joints.LFJ3,
        Joints.LFJ2,
        Joints.LFJ1,
        Joints.LFJ0,
    ),
    # Thumb has 5 actuators.
    Components.TH: (
        Joints.THJ4,
        Joints.THJ3,
        Joints.THJ2,
        Joints.THJ1,
        Joints.THJ0,
    ),
}

# ====================== #
# Actuation constants
# ====================== #


class Actuators(enum.Enum):
    """Actuators of the Shadow Hand."""

    # Wrist.
    A_WRJ1 = enum.auto()  # Horizontal movement.
    A_WRJ0 = enum.auto()  # Vertical movement.
    # First finger.
    A_FFJ3 = enum.auto()  # Horizontal movement.
    A_FFJ2 = enum.auto()  # Vertical movement.
    A_FFJ1 = enum.auto()  # Vertical movement, bending coupled joints.
    # Middle finger.
    A_MFJ3 = enum.auto()  # Horizontal movement.
    A_MFJ2 = enum.auto()  # Vertical movement.
    A_MFJ1 = enum.auto()  # Vertical movement, bending coupled joints.
    # Ring finger.
    A_RFJ3 = enum.auto()  # Horizontal movement.
    A_RFJ2 = enum.auto()  # Vertical movement.
    A_RFJ1 = enum.auto()  # Vertical movement, bending coupled joints.
    # Little finger.
    A_LFJ4 = enum.auto()  # Vertical movement, towards center of palm.
    A_LFJ3 = enum.auto()  # Horizontal movement.
    A_LFJ2 = enum.auto()  # Vertical movement.
    A_LFJ1 = enum.auto()  # Vertical movement, bending coupled joints.
    # Thumb.
    A_THJ4 = enum.auto()  # Rotational movement of the thumb.
    A_THJ3 = enum.auto()  # Bending movement.
    A_THJ2 = enum.auto()  # Bending movement.
    A_THJ1 = enum.auto()  # Bending movement.
    A_THJ0 = enum.auto()  # Bending movement.


# The total number of actuators.
NUM_ACTUATORS: int = len(Actuators)

# A list of actuator names, as strings.
ACTUATOR_NAMES: List[str] = [a.name for a in Actuators]

# A mapping from `Components` to the list of `Actuators` that belong to it.
ACTUATOR_GROUP: Dict[Components, Tuple[Actuators, ...]] = {
    # Wrist has 2 actuators.
    Components.WR: (Actuators.A_WRJ1, Actuators.A_WRJ0),
    # First, middle and ring fingers have 3 actuators.
    Components.FF: (Actuators.A_FFJ3, Actuators.A_FFJ2, Actuators.A_FFJ1),
    Components.MF: (Actuators.A_MFJ3, Actuators.A_MFJ2, Actuators.A_MFJ1),
    Components.RF: (Actuators.A_RFJ3, Actuators.A_RFJ2, Actuators.A_RFJ1),
    # Little finger has 4 actuators.
    Components.LF: (
        Actuators.A_LFJ4,
        Actuators.A_LFJ3,
        Actuators.A_LFJ2,
        Actuators.A_LFJ1,
    ),
    # Thumb has 5 actuators.
    Components.TH: (
        Actuators.A_THJ4,
        Actuators.A_THJ3,
        Actuators.A_THJ2,
        Actuators.A_THJ1,
        Actuators.A_THJ0,
    ),
}

# One-to-many mapping from `Actuators` to the joint(s) it controls.
# The first two joints of each of the main fingers are coupled, which means there is
# only one actuator controlling them via a single tendon.
ACTUATOR_JOINT_MAPPING: Dict[Actuators, Tuple[Joints, ...]] = {
    # Wrist.
    Actuators.A_WRJ1: (Joints.WRJ1,),
    Actuators.A_WRJ0: (Joints.WRJ0,),
    # First finger.
    Actuators.A_FFJ3: (Joints.FFJ3,),
    Actuators.A_FFJ2: (Joints.FFJ2,),
    Actuators.A_FFJ1: (Joints.FFJ1, Joints.FFJ0),
    # Middle finger.
    Actuators.A_MFJ3: (Joints.MFJ3,),
    Actuators.A_MFJ2: (Joints.MFJ2,),
    Actuators.A_MFJ1: (Joints.MFJ1, Joints.MFJ0),
    # Ring finger.
    Actuators.A_RFJ3: (Joints.RFJ3,),
    Actuators.A_RFJ2: (Joints.RFJ2,),
    Actuators.A_RFJ1: (Joints.RFJ1, Joints.RFJ0),
    # Little finger.
    Actuators.A_LFJ4: (Joints.LFJ4,),
    Actuators.A_LFJ3: (Joints.LFJ3,),
    Actuators.A_LFJ2: (Joints.LFJ2,),
    Actuators.A_LFJ1: (Joints.LFJ1, Joints.LFJ0),
    # Thumb.
    Actuators.A_THJ4: (Joints.THJ4,),
    Actuators.A_THJ3: (Joints.THJ3,),
    Actuators.A_THJ2: (Joints.THJ2,),
    Actuators.A_THJ1: (Joints.THJ1,),
    Actuators.A_THJ0: (Joints.THJ0,),
}

# Reverse mapping of `ACTUATOR_JOINT_MAPPING`.
JOINT_ACTUATOR_MAPPING: Dict[Joints, Actuators] = {
    v: k for k, vs in ACTUATOR_JOINT_MAPPING.items() for v in vs
}


def _compute_projection_matrices() -> Tuple[np.ndarray, np.ndarray, List[List[int]]]:
    position_to_control = np.zeros((NUM_ACTUATORS, NUM_JOINTS))
    control_to_position = np.zeros((NUM_JOINTS, NUM_ACTUATORS))
    coupled_joint_ids = []
    actuator_ids = dict(zip(Actuators, range(NUM_ACTUATORS)))
    joint_ids = dict(zip(Joints, range(NUM_JOINTS)))
    for actuator, joints in ACTUATOR_JOINT_MAPPING.items():
        value = 1.0 / len(joints)
        a_id = actuator_ids[actuator]
        j_ids = np.array([joint_ids[joint] for joint in joints])
        if len(joints) > 1:
            coupled_joint_ids.append([joint_ids[joint] for joint in joints])
        position_to_control[a_id, j_ids] = 1.0
        control_to_position[j_ids, a_id] = value
    return position_to_control, control_to_position, coupled_joint_ids


# Projection matrices for mapping control space to joint space and vice versa. These
# matrices should premultiply the vector to be projected.
# POSITION_TO_CONTROL maps a control vector to a joint vector.
# CONTROL_TO_POSITION maps a joint vector to a control vector.
(
    POSITION_TO_CONTROL,
    CONTROL_TO_POSITION,
    COUPLED_JOINT_IDS,
) = _compute_projection_matrices()

# ====================== #
# Limits
# ====================== #

# Note: It seems there's a discrepancy between the values reported in the spec sheet^[1]
# and the values reported in the company's github repo^[2]. I'm going to follow the ones
# on the spec sheet, since it seems those are the ones OpenAI^[3] used for their project
# as well.
# References:
#   [1]: Shadow Robot spec sheet: https://www.shadowrobot.com/wp-content/uploads/shadow_dexterous_hand_technical_specification_E_20190221.pdf
#   [2]: Shadow Robot Company code: github.com/shadow-robot/sr_common
#   [3]: OpenAI code: github.com/openai/robogym
#
# A mapping from `Actuators` to the corresponding control range, in radians.
ACTUATOR_CTRLRANGE: Dict[Actuators, Tuple[float, float]] = {
    # Wrist.
    Actuators.A_WRJ1: (rad(-28), rad(8)),
    Actuators.A_WRJ0: (rad(-40), rad(28)),
    # First finger.
    Actuators.A_FFJ3: (rad(-20), rad(20)),
    Actuators.A_FFJ2: (rad(0), rad(90)),
    Actuators.A_FFJ1: (rad(0), rad(180)),
    # Middle finger.
    Actuators.A_MFJ3: (rad(-20), rad(20)),
    Actuators.A_MFJ2: (rad(0), rad(90)),
    Actuators.A_MFJ1: (rad(0), rad(180)),
    # Ring finger.
    Actuators.A_RFJ3: (rad(-20), rad(20)),
    Actuators.A_RFJ2: (rad(0), rad(90)),
    Actuators.A_RFJ1: (rad(0), rad(180)),
    # Little finger.
    Actuators.A_LFJ4: (rad(0), rad(45)),
    Actuators.A_LFJ3: (rad(-20), rad(20)),
    Actuators.A_LFJ2: (rad(0), rad(90)),
    Actuators.A_LFJ1: (rad(0), rad(180)),
    # Thumb.
    Actuators.A_THJ4: (rad(-60), rad(60)),
    Actuators.A_THJ3: (rad(0), rad(70)),
    Actuators.A_THJ2: (rad(-12), rad(12)),
    Actuators.A_THJ1: (rad(-30), rad(30)),
    # TODO(kevin): OpenAI uses (-90, 0) here, figure out why.
    Actuators.A_THJ0: (rad(0), rad(90)),
}

# Joint position limits, in radians.
# Coupled joints share the full ctrlrange, so their range is split in half.
# Note: These values match the values reported in the spec sheet^[1], page 7.
JOINT_LIMITS: Dict[Joints, Tuple[float, float]] = {}
for actuator, ctrlrange in ACTUATOR_CTRLRANGE.items():
    joints = ACTUATOR_JOINT_MAPPING[actuator]
    for joint in joints:
        JOINT_LIMITS[joint] = (
            ctrlrange[0] / len(joints),
            ctrlrange[1] / len(joints),
        )

# Effort limits.
# For a `hinge` (revolute) joint, this is equivalent to the torque limit, in N-m.
# For a `slide` (prismatic) joint, this is equivalent to the force limit, in N.
# Taken from company's github repo^[2] by parsing the XACRO files.
EFFORT_LIMITS: Dict[Actuators, Tuple[float, float]] = {
    # Wrist.
    Actuators.A_WRJ1: (-10.0, 10.0),
    Actuators.A_WRJ0: (-30.0, 30.0),
    # First finger.
    Actuators.A_FFJ3: (-2.0, 2.0),
    Actuators.A_FFJ2: (-2.0, 2.0),
    Actuators.A_FFJ1: (-2.0, 2.0),
    # Middle finger.
    Actuators.A_MFJ3: (-2.0, 2.0),
    Actuators.A_MFJ2: (-2.0, 2.0),
    Actuators.A_MFJ1: (-2.0, 2.0),
    # Ring finger.
    Actuators.A_RFJ3: (-2.0, 2.0),
    Actuators.A_RFJ2: (-2.0, 2.0),
    Actuators.A_RFJ1: (-2.0, 2.0),
    # Little finger.
    Actuators.A_LFJ4: (-2.0, 2.0),
    Actuators.A_LFJ3: (-2.0, 2.0),
    Actuators.A_LFJ2: (-2.0, 2.0),
    Actuators.A_LFJ1: (-2.0, 2.0),
    # Thumb.
    Actuators.A_THJ4: (-2.0, 2.0),
    Actuators.A_THJ3: (-2.0, 2.0),
    Actuators.A_THJ2: (-2.0, 2.0),
    Actuators.A_THJ1: (-2.0, 2.0),
    Actuators.A_THJ0: (-2.0, 2.0),
}

# Joint velocity limits, in rad/s.
# Taken from company's github repo^[2] by parsing the XACRO files.
# NOTE(kevin): It seems all the actuators have the same velocity limits.
VELOCITY_LIMITS: Dict[Actuators, Tuple[float, float]] = {
    # Wrist.
    Actuators.A_WRJ1: (-2.0, 2.0),
    Actuators.A_WRJ0: (-2.0, 2.0),
    # First finger.
    Actuators.A_FFJ3: (-2.0, 2.0),
    Actuators.A_FFJ2: (-2.0, 2.0),
    Actuators.A_FFJ1: (-2.0, 2.0),
    # Middle finger.
    Actuators.A_MFJ3: (-2.0, 2.0),
    Actuators.A_MFJ2: (-2.0, 2.0),
    Actuators.A_MFJ1: (-2.0, 2.0),
    # Ring finger.
    Actuators.A_RFJ3: (-2.0, 2.0),
    Actuators.A_RFJ2: (-2.0, 2.0),
    Actuators.A_RFJ1: (-2.0, 2.0),
    # Little finger.
    Actuators.A_LFJ4: (-2.0, 2.0),
    Actuators.A_LFJ3: (-2.0, 2.0),
    Actuators.A_LFJ2: (-2.0, 2.0),
    Actuators.A_LFJ1: (-2.0, 2.0),
    # Thumb.
    Actuators.A_THJ4: (-2.0, 2.0),
    Actuators.A_THJ3: (-2.0, 2.0),
    Actuators.A_THJ2: (-2.0, 2.0),
    Actuators.A_THJ1: (-2.0, 2.0),
    Actuators.A_THJ0: (-2.0, 2.0),
}

# ====================== #
# Tendon constants
# ====================== #


class Tendons(enum.Enum):
    """Tendons of the Shadow Hand.

    These are used to model the underactuation of the *FJ0 and *FJ1 joints of the main
    fingers. A tendon is defined for each *FJ0-*FJ1 pair, and an actuator is used to
    drive it.
    """

    FFT1 = enum.auto()  # First finger.
    MFT1 = enum.auto()  # Middle finger.
    RFT1 = enum.auto()  # Ring finger.
    LFT1 = enum.auto()  # Little finger.


# The total number of tendons.
NUM_TENDONS: int = len(Tendons)

# A list of tendon names, as strings.
TENDON_NAMES: List[str] = [t.name for t in Tendons]

# Mapping from `Tendons` to `Joints` pair.
TENDON_JOINT_MAPPING: Dict[Tendons, Tuple[Joints, Joints]] = {
    Tendons.FFT1: (Joints.FFJ0, Joints.FFJ1),  # First finger.
    Tendons.MFT1: (Joints.MFJ0, Joints.MFJ1),  # Middle finger.
    Tendons.RFT1: (Joints.RFJ0, Joints.RFJ1),  # Ring finger.
    Tendons.LFT1: (Joints.LFJ0, Joints.LFJ1),  # Little finger.
}

# Mapping from `Tendons` to the `Actuators` that drives it.
TENDON_ACTUATOR_MAPPING: Dict[Tendons, Actuators] = {
    Tendons.FFT1: Actuators.A_FFJ1,  # First finger.
    Tendons.MFT1: Actuators.A_MFJ1,  # Middle finger.
    Tendons.RFT1: Actuators.A_RFJ1,  # Ring finger.
    Tendons.LFT1: Actuators.A_LFJ1,  # Little finger.
}
# Reverse mapping of `TENDON_ACTUATOR_MAPPING`.
ACTUATOR_TENDON_MAPPING: Dict[Actuators, Tendons] = {
    v: k for k, v in TENDON_ACTUATOR_MAPPING.items()
}

# ====================== #
# Fingertip constants
# ====================== #

FINGERTIP_NAMES: Tuple[str, ...] = (
    "fftip",
    "mftip",
    "rftip",
    "lftip",
    "thtip",
)

# Mapping from finger `Components` to its associated fingertip body name.
FINGER_FINGERTIP_MAPPING: Dict[Components, str] = {
    Components.FF: "fftip",
    Components.MF: "mftip",
    Components.RF: "rftip",
    Components.LF: "lftip",
    Components.TH: "thtip",
}

# ====================== #
# Other constants
# ====================== #

# Names of the <geom> tags in the XML file whose color can be changed. This is useful
# for dynamically changing the colors of the hand components.
COLORED_GEOMS: Tuple[str, ...] = (
    "forearm",
    "wrist",
    "palm",
    "ffproximal",
    "ffmiddle",
    "ffdistal",
    "mfproximal",
    "mfmiddle",
    "mfdistal",
    "rfproximal",
    "rfmiddle",
    "rfdistal",
    "lfmetacarpal",
    "lfproximal",
    "lfmiddle",
    "lfdistal",
    "thproximal",
    "thmiddle",
    "thdistal",
)

# Mapping from finger `Components` to its associated geoms.
FINGER_GEOM_MAPPING: Dict[Components, Tuple[str, ...]] = {
    Components.FF: ("ffproximal", "ffmiddle", "ffdistal"),
    Components.MF: ("mfproximal", "mfmiddle", "mfdistal"),
    Components.RF: ("rfproximal", "rfmiddle", "rfdistal"),
    Components.LF: ("lfproximal", "lfmiddle", "lfdistal"),
    Components.TH: ("thproximal", "thmiddle", "thdistal"),
}



from __future__ import annotations

import abc
import dataclasses
from typing import Optional, Sequence

import mujoco
import numpy as np

from dexterity import hints


class CartesianVelocitytoJointVelocityMapper(abc.ABC):
    """Abstract base class for a Cartesian velocity to joint velocity mapper.

    Subclasses should implement the `compute_joint_velocities` method, which maps a
    Cartesian 6D velocity about the global frame to joint velocities. The target
    Cartesian velocity is specified as a 6D vector, with the 3D linear velocity term
    followed by the 3D angular velocity term.

    At every call to `compute_joint_velocities`, the mapper will attempt to solve for
    the joint velocities that realize the target Cartesian velocity on the frame
    attached to the MuJoCo object defined by `object_type` and `object_name`. The target
    Cartesian velocity must be expressed about the MuJoCo object's origin, in world
    orientation. An error is returned if the computed velocities do not exist, or if the
    mapper fails to compute the velocities.

    This class expects an updated MjData object at every call to
    `compute_joint_velocities`, so users should ensure that the mjData object has
    up-to-date `qpos` and `qvel` fields.
    """

    @abc.abstractmethod
    def compute_joint_velocities(
        self,
        data: hints.MjData,
        target_velocities: Sequence[np.ndarray],
        nullspace_bias: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Maps the Cartesian target velocity to joint velocities."""


@dataclasses.dataclass
class Parameters:
    """Container for `CartesianVelocitytoJointVelocityMapper` parameters."""

    model: hints.MjModel
    """MuJoCo `MjModel` instance."""

    object_types: Sequence[hints.MujocoObjectType]
    """MuJoCo type of the object being controlled. Only bodies, geoms and sites are
    currently supported."""

    object_names: Sequence[str]
    """MuJoCo name of the object being controlled."""

    def validate_parameters(self) -> None:
        """Validates the parameters."""

        for object_type in self.object_types:
            if object_type not in [
                mujoco.mjtObj.mjOBJ_BODY,
                mujoco.mjtObj.mjOBJ_GEOM,
                mujoco.mjtObj.mjOBJ_SITE,
            ]:
                raise ValueError(
                    f"Objects of type {object_type} are not supported. Only"
                    " bodies, geoms and sites are supported."
                )

        for object_name, object_type in zip(self.object_names, self.object_types):
            if self.model.name2id(object_name, object_type) < 0:
                raise ValueError(
                    f"Could not find MuJoCo object with name {object_name} and"
                    f" type {object_type} in the provided model."
                )



import dm_env
from absl import logging
from dm_control import composer

from dexterity import exception
from dexterity import task


class GoalEnvironment(composer.Environment):
    """A composer.Environment for `GoalTask`s."""

    task: task.GoalTask

    def reset(self) -> dm_env.TimeStep:
        failed_attempts = 0
        while True:
            try:
                return super().reset()
            except exception.GoalInitializationError as e:
                failed_attempts += 1
                logging.error(
                    "Error %d during episode reset: %s", failed_attempts, repr(e)
                )

    def step(self, action) -> dm_env.TimeStep:
        failed_attempts = 0
        while True:
            try:
                return super().step(action)
            except exception.GoalInitializationError as e:
                failed_attempts += 1
                logging.error(
                    "Error %d during episode step: %s", failed_attempts, repr(e)
                )



"""Tests for manipulation loader."""

from absl.testing import absltest

from dexterity import manipulation


class LoaderConstantsTest(absltest.TestCase):
    def testConstants(self) -> None:
        self.assertNotEmpty(manipulation.ALL_TASKS)
        self.assertNotEmpty(manipulation.ALL_NAMES)
        self.assertNotEmpty(manipulation.TASKS_BY_DOMAIN)


if __name__ == "__main__":
    absltest.main()



from typing import Optional

from dexterity import arena
from dexterity import hints


class Standard(arena.Arena):
    """Subclass of the standard Composer arena for the manipulation suite of tasks.

    Has a checkered ground plane.
    """

    def _build(self, name: Optional[str] = None) -> None:
        super()._build(name=name)

        # Remove the default visual headlight setting in `dm_control.Arena`.
        self._mjcf_root.visual.remove("headlight")

        # Add visual assets.
        self.mjcf_model.asset.add(
            "texture",
            type="skybox",
            builtin="gradient",
            rgb1=(0.4, 0.6, 0.8),
            rgb2=(0.0, 0.0, 0.0),
            width=100,
            height=100,
        )
        groundplane_texture = self.mjcf_model.asset.add(
            "texture",
            name="groundplane",
            type="2d",
            builtin="checker",
            rgb1=(0.2, 0.3, 0.4),
            rgb2=(0.1, 0.2, 0.3),
            width=300,
            height=300,
            mark="edge",
            markrgb=(0.8, 0.8, 0.8),
        )
        groundplane_material = self.mjcf_model.asset.add(
            "material",
            name="groundplane",
            texture=groundplane_texture,
            texrepeat=(5, 5),
            texuniform="true",
            reflectance=0.2,
        )

        # Add ground plane.
        self._ground = self.mjcf_model.worldbody.add(
            "geom",
            name="ground",
            type="plane",
            material=groundplane_material,
            size=(1, 1, 0.1),
            friction=(0.4,),
            solimp=(0.95, 0.99, 0.001),
            solref=(0.002, 1),
        )

        # Add lighting.
        self.mjcf_model.worldbody.add(
            "light",
            pos=(0, 0, 1.5),
            dir=(0, 0, -1),
            diffuse=(0.7, 0.7, 0.7),
            specular=(0.3, 0.3, 0.3),
            directional="true",
            castshadow="true",
        )

        # Always initialize the free camera so that it points at the origin.
        self.mjcf_model.statistic.center = (0.0, 0.0, 0.0)

    @property
    def ground(self) -> hints.MjcfElement:
        """The ground plane mjcf element."""
        return self._ground



"""Tools for defining and visualizing workspaces for in-hand manipulation tasks."""

import dataclasses
from typing import Sequence, Tuple

import numpy as np

from dexterity import hints
from dexterity.manipulation.shared import constants

# Ensures that all site dimensions are positive.
_MIN_SITE_DIMENSION = 1e-6


@dataclasses.dataclass(frozen=True)
class BoundingBox:
    lower: Tuple[float, ...]
    upper: Tuple[float, ...]


def add_bbox_site(
    body: hints.MjcfElement,
    lower: Sequence[float],
    upper: Sequence[float],
    visible: bool = False,
    **kwargs,
) -> hints.MjcfElement:
    """Adds a site for visualizing a bounding box to an MJCF model."""
    assert len(lower) == len(upper) == 3
    lower_arr = np.array(lower)
    upper_arr = np.array(upper)
    assert np.all(lower_arr <= upper_arr)
    pos = (upper_arr + lower_arr) / 2.0
    size = np.maximum((upper_arr - lower_arr) / 2.0, _MIN_SITE_DIMENSION)
    group = None if visible else constants.TASK_SITE_GROUP
    return body.add(
        "site",
        type="box",
        pos=pos,
        size=size,
        group=group,
        **kwargs,
    )


def add_target_site(
    body: hints.MjcfElement,
    radius: float,
    visible: bool = False,
    **kwargs,
) -> hints.MjcfElement:
    """Adds a site for visualizing a target location to an MJCF model."""
    assert radius > 0.0
    group = None if visible else constants.TASK_SITE_GROUP
    return body.add(
        "site",
        type="sphere",
        size=[radius],
        group=group,
        **kwargs,
    )



"""Shadow hand actuation."""

import dataclasses
import enum
from typing import Dict, Tuple

from dexterity.models.hands import shadow_hand_e_constants as consts


class Actuation(enum.Enum):
    """Available actuation methods for the hand."""

    POSITION = enum.auto()
    """In this mode, position actuators are created for the joints and tendons of the
    hand. Incoming control commands are interepreted as desired joint positions and the
    underlying position controller chooses the right force for each actuator to reach
    these positions.
    """


@dataclasses.dataclass(frozen=True)
class ActuatorParams:
    """PD parameters for the position actuators of the hand."""

    kp: float
    damping: float = 1.0


# Actuation limits for the hand.
ACTUATION_LIMITS: Dict[Actuation, Dict[consts.Actuators, Tuple[float, float]]] = {
    # For position control, actuation limits are joint limits.
    Actuation.POSITION: consts.ACTUATOR_CTRLRANGE,
}


_WR_DAMPING = 1.0
_FJ4_DAMPING = 0.1
_FJ3_DAMPING = 0.1
_FJ2_DAMPING = 0.1
_FJ1_DAMPING = 0.1
_TH_DAMPING = 0.2

_WR_GAIN = 20.0
_TH_GAIN = 3.0
_FJ4_GAIN = 2.0
_FJ3_GAIN = 2.0
_FJ2_GAIN = 2.0
_FJ1_GAIN = 0.8

ACTUATOR_PARAMS: Dict[Actuation, Dict[consts.Actuators, ActuatorParams]] = {
    Actuation.POSITION: {
        # Wrist.
        consts.Actuators.A_WRJ1: ActuatorParams(kp=_WR_GAIN, damping=_WR_DAMPING),
        consts.Actuators.A_WRJ0: ActuatorParams(kp=_WR_GAIN, damping=_WR_DAMPING),
        # First finger.
        consts.Actuators.A_FFJ3: ActuatorParams(kp=_FJ3_GAIN, damping=_FJ3_DAMPING),
        consts.Actuators.A_FFJ2: ActuatorParams(kp=_FJ2_GAIN, damping=_FJ2_DAMPING),
        consts.Actuators.A_FFJ1: ActuatorParams(kp=_FJ1_GAIN, damping=_FJ1_DAMPING),
        # Middle finger.
        consts.Actuators.A_MFJ3: ActuatorParams(kp=_FJ3_GAIN, damping=_FJ3_DAMPING),
        consts.Actuators.A_MFJ2: ActuatorParams(kp=_FJ2_GAIN, damping=_FJ2_DAMPING),
        consts.Actuators.A_MFJ1: ActuatorParams(kp=_FJ1_GAIN, damping=_FJ1_DAMPING),
        # Ring finger.
        consts.Actuators.A_RFJ3: ActuatorParams(kp=_FJ3_GAIN, damping=_FJ3_DAMPING),
        consts.Actuators.A_RFJ2: ActuatorParams(kp=_FJ2_GAIN, damping=_FJ2_DAMPING),
        consts.Actuators.A_RFJ1: ActuatorParams(kp=_FJ1_GAIN, damping=_FJ1_DAMPING),
        # Little finger.
        consts.Actuators.A_LFJ4: ActuatorParams(kp=_FJ4_GAIN, damping=_FJ4_DAMPING),
        consts.Actuators.A_LFJ3: ActuatorParams(kp=_FJ3_GAIN, damping=_FJ3_DAMPING),
        consts.Actuators.A_LFJ2: ActuatorParams(kp=_FJ2_GAIN, damping=_FJ2_DAMPING),
        consts.Actuators.A_LFJ1: ActuatorParams(kp=_FJ1_GAIN, damping=_FJ1_DAMPING),
        # Thumb.
        consts.Actuators.A_THJ4: ActuatorParams(kp=_TH_GAIN, damping=_TH_DAMPING),
        consts.Actuators.A_THJ3: ActuatorParams(kp=_TH_GAIN, damping=_TH_DAMPING),
        consts.Actuators.A_THJ2: ActuatorParams(kp=_TH_GAIN, damping=_TH_DAMPING),
        consts.Actuators.A_THJ1: ActuatorParams(kp=_TH_GAIN, damping=_TH_DAMPING),
        consts.Actuators.A_THJ0: ActuatorParams(kp=_TH_GAIN, damping=_TH_DAMPING),
    },
}



"""Tests for hand_effector."""

import numpy as np
from absl.testing import absltest
from dm_control import mjcf

from dexterity.effectors import hand_effector
from dexterity.models.hands import shadow_hand_e


class HandEffectorTest(absltest.TestCase):
    def test_set_control(self) -> None:
        hand = shadow_hand_e.ShadowHandSeriesE()
        effector = hand_effector.HandEffector(hand=hand, hand_name=hand.name)
        physics = mjcf.Physics.from_mjcf_model(hand.mjcf_model)
        action_spec = effector.action_spec(physics)
        rand_ctrl = np.random.uniform(*physics.bind(hand.actuators).ctrlrange.T)
        rand_ctrl = rand_ctrl.astype(action_spec.dtype)
        effector.set_control(physics, rand_ctrl)
        np.testing.assert_allclose(physics.bind(hand.actuators).ctrl, rand_ctrl)


if __name__ == "__main__":
    absltest.main()



from typing import Dict, List

import numpy as np
from dm_control import composer
from dm_control import mjcf

from dexterity.hints import MjcfElement
from dexterity.models.hands import dexterous_hand
from dexterity.models.hands import shadow_hand_e_actuation as sh_actuation
from dexterity.models.hands import shadow_hand_e_constants as consts
from dexterity.utils import mujoco_actuation
from dexterity.utils import mujoco_utils


class ShadowHandSeriesE(dexterous_hand.DexterousHand):
    """Shadow Dexterous Hand E Series."""

    def _build(
        self,
        name: str = "shadow_hand_e",
        actuation: sh_actuation.Actuation = sh_actuation.Actuation.POSITION,
    ) -> None:
        """Initializes the hand.

        Args:
            name: The name of the hand. Used as a prefix in the MJCF name attributes.
            actuation: Instance of `shadow_hand_e_actuation.Actuation` specifying which
                actuation method to use.
        """
        self._mjcf_root = mjcf.from_path(str(consts.SHADOW_HAND_E_XML))
        self._mjcf_root.model = name
        self._actuation = actuation

        self._parse_mjcf_elements()
        self._add_fingertip_sites()
        self._add_tendons()
        self._add_actuators()
        self._add_sensors()

    def initialize_episode(
        self, physics: mjcf.Physics, random_state: np.random.RandomState
    ) -> None:
        del random_state  # Unused.

        # Apply gravity compensation.
        mujoco_utils.compensate_gravity(physics, self.mjcf_model.find_all("body"))

    # ================= #
    # Accessors.
    # ================= #

    @property
    def mjcf_model(self) -> mjcf.RootElement:
        return self._mjcf_root

    @property
    def name(self) -> str:
        return self._mjcf_root.model

    @property
    def joints(self) -> List[MjcfElement]:
        """List of joint elements belonging to the hand."""
        return self._joints

    @property
    def actuators(self) -> List[MjcfElement]:
        """List of actuator elements belonging to the hand."""
        return self._actuators

    @property
    def tendons(self) -> List[MjcfElement]:
        """List of tendon elements belonging to the hand."""
        return self._tendons

    @property
    def joint_torque_sensors(self) -> List[MjcfElement]:
        """List of joint torque sensor elements belonging to the hand."""
        return self._joint_torque_sensors

    @property
    def fingertip_sites(self) -> List[MjcfElement]:
        """List of fingertip site elements belonging to the hand."""
        return self._fingertip_sites

    # ================= #
    # Public methods.
    # ================= #

    def control_to_joint_positions(self, control: np.ndarray) -> np.ndarray:
        """Maps a 20-D position control command to a 24-D joint position command.

        The control commands for the coupled joints are evenly split amongst them.
        """
        if control.shape != (self._num_actuators,):
            raise ValueError(
                f"Expected control of shape ({self._num_actuators}), got"
                f" {control.shape}"
            )
        return consts.CONTROL_TO_POSITION @ control

    def joint_positions_to_control(self, qpos: np.ndarray) -> np.ndarray:
        """Maps a 24-D joint position command to a 20-D control command.

        The position commands for the coupled joints are summed up to form the control
        for their corresponding actuator.
        """
        if qpos.shape != (self._num_joints,):
            raise ValueError(
                f"Expected qpos of shape ({self._num_joints}), got {qpos.shape}"
            )
        return consts.POSITION_TO_CONTROL @ qpos

    def set_joint_angles(self, physics: mjcf.Physics, joint_angles: np.ndarray) -> None:
        """Sets the joints of the hand to a given configuration."""
        physics.bind(self._joints).qpos = joint_angles

    def sample_joint_angles(
        self, physics: mjcf.Physics, random_state: np.random.RandomState
    ) -> np.ndarray:
        qpos = random_state.uniform(*physics.bind(self._joints).range.T)

        # Ensure coupled joints have the same joint values.
        for coupled_ids in consts.COUPLED_JOINT_IDS:
            val = qpos[coupled_ids[-1]]
            qpos[coupled_ids] = val

        return qpos

    # ================= #
    # Private methods.
    # ================= #

    def _parse_mjcf_elements(self) -> None:
        """Parses MJCF elements that will be exposed as attributes."""
        # Parse joints.
        self._joints: List[mjcf.Element] = []
        self._joint_elem_mapping: Dict[consts.Joints, mjcf.Element] = {}
        for joint in consts.Joints:
            joint_elem = self._mjcf_root.find("joint", joint.name)
            if joint_elem is None:
                raise ValueError(f"Could not find joint {joint.name} in MJCF model.")
            self._joints.append(joint_elem)
            self._joint_elem_mapping[joint] = joint_elem
        self._num_joints = len(self._joints)

    def _add_fingertip_sites(self) -> None:
        """Adds sites to the tips of the fingers of the hand."""
        self._fingertip_sites: List[mjcf.Element] = []
        self._fingertip_site_elem_mapping: Dict[consts.Components, mjcf.Element] = {}
        for finger, tip_name in consts.FINGER_FINGERTIP_MAPPING.items():
            tip_elem = self._mjcf_root.find("body", tip_name)
            if tip_elem is None:
                raise ValueError(f"Could not find fingertip {tip_name} in MJCF model.")
            tip_site = tip_elem.add(
                "site",
                name=tip_name + "_site",
                pos="0 0 0",
                # NOTE(kevin): The kwargs below are for visualization purposes.
                size="0.001 0.001 0.001",
                type="sphere",
                rgba="1 0 0 1",
                group=composer.SENSOR_SITES_GROUP,
            )
            self._fingertip_sites.append(tip_site)
            self._fingertip_site_elem_mapping[finger] = tip_site

    def _add_tendons(self) -> None:
        """Add tendons to the hand."""
        self._tendons: List[mjcf.Element] = []
        self._tendon_elem_mapping: Dict[consts.Tendons, mjcf.Element] = {}
        for tendon, joints in consts.TENDON_JOINT_MAPPING.items():
            tendon_elem = self._mjcf_root.tendon.add("fixed", name=tendon.name)
            for joint in joints:
                # We set `coef=1` to make the tendon's length equal to the sum of the
                # joint positions.
                tendon_elem.add("joint", joint=joint.name, coef=1.0)
            self._tendons.append(tendon_elem)
            self._tendon_elem_mapping[tendon] = tendon_elem

    def _add_actuators(self) -> None:
        """Adds actuators to the hand."""
        if self._actuation not in sh_actuation.Actuation:
            raise ValueError(
                f"Actuation {self._actuation} is not a valid actuation mode."
            )

        if self._actuation == sh_actuation.Actuation.POSITION:
            self._add_position_actuators()

    def _add_position_actuators(self) -> None:
        """Adds position actuators to the mjcf model."""

        self._mjcf_root.default.general.forcelimited = "true"
        self._mjcf_root.actuator.motor.clear()

        self._actuators: List[mjcf.Element] = []
        self._actuator_elem_mapping: Dict[consts.Actuators, mjcf.Element] = {}
        for actuator, actuator_params in sh_actuation.ACTUATOR_PARAMS[
            self._actuation
        ].items():
            if actuator in consts.ACTUATOR_TENDON_MAPPING:
                elem = self._tendon_elem_mapping[
                    consts.ACTUATOR_TENDON_MAPPING[actuator]
                ]
                elem_type = "tendon"
            else:
                elem = self._joint_elem_mapping[
                    consts.ACTUATOR_JOINT_MAPPING[actuator][0]
                ]
                elem_type = "joint"
                elem.damping = actuator_params.damping

            qposrange = sh_actuation.ACTUATION_LIMITS[self._actuation][actuator]
            actuator_elem = mujoco_actuation.add_position_actuator(
                elem=elem,
                elem_type=elem_type,
                qposrange=qposrange,
                ctrlrange=qposrange,
                kp=actuator_params.kp,
                forcerange=consts.EFFORT_LIMITS[actuator],
                name=actuator.name,
            )

            self._actuator_elem_mapping[actuator] = actuator_elem
            self._actuators.append(actuator_elem)

        self._num_actuators = len(self._actuators)

    def _add_sensors(self) -> None:
        """Add sensors to the mjcf model."""
        self._add_torque_sensors()

    def _add_torque_sensors(self) -> None:
        """Adds torque sensors to the joints of the hand."""
        self._joint_torque_sensors = []
        self._joint_torque_sensor_elem_mapping = {}
        for joint in consts.Joints:
            joint_elem = self._joint_elem_mapping[joint]
            site_elem = joint_elem.parent.add(
                "site",
                name=joint_elem.name + "_site",
                # NOTE(kevin): The kwargs below are for visualization purposes.
                size="0.001 0.001 0.001",
                type="box",
                rgba="0 1 0 1",
                group=composer.SENSOR_SITES_GROUP,
            )
            # Create a 3-axis torque sensor.
            torque_sensor_elem = joint_elem.root.sensor.add(
                "torque",
                site=site_elem,
                name=joint_elem.name + "_torque",
            )
            self._joint_torque_sensors.append(torque_sensor_elem)
            self._joint_torque_sensor_elem_mapping[joint] = torque_sensor_elem



from dexterity.manipulation.props.juggling_ball import JugglingBall
from dexterity.manipulation.props.openai_cube import OpenAICube
from dexterity.manipulation.props.target_sphere import TargetSphere

__all__ = [
    "OpenAICube",
    "TargetSphere",
    "JugglingBall",
]



from dm_control.entities.props import primitive

from dexterity import _SRC_ROOT

_TEXTURE_PATH = _SRC_ROOT / "manipulation" / "props" / "juggling_ball" / "rgby.png"


class JugglingBall(primitive.Primitive):
    def _build(
        self,
        radius: float = 0.01,
        name: str = "ball",
    ) -> None:
        """Builds the ball."""

        super()._build(
            geom_type="sphere",
            size=[radius] * 3,
            condim=6,
            friction="1 .001 .001",
            name=name,
        )

        self.mjcf_model.asset.add(
            "texture",
            name="texture_ball",
            file=str(_TEXTURE_PATH),
        )
        self.mjcf_model.asset.add(
            "material",
            name="texture_ball",
            texture="texture_ball",
            specular="1",
            shininess=".3",
            reflectance="0.0",
            rgba="1 1 1 1",
        )

        setattr(self.mjcf_model.find("geom", "geom"), "material", "texture_ball")

    @property
    def name(self) -> str:
        return self.mjcf_model.model



from typing import Any

import dm_env


class Wrapper(dm_env.Environment):
    """Base class for composer.Environment wrappers."""

    def __init__(self, environment: dm_env.Environment):
        self._environment = environment

    def __getattr__(self, name: str) -> Any:
        return getattr(self._environment, name)

    @property
    def environment(self) -> dm_env.Environment:
        return self._environment

    def step(self, action) -> dm_env.TimeStep:
        return self._environment.step(action)

    def reset(self) -> dm_env.TimeStep:
        return self._environment.reset()

    def action_spec(self):
        return self._environment.action_spec()

    def discount_spec(self):
        return self._environment.discount_spec()

    def observation_spec(self):
        return self._environment.observation_spec()

    def reward_spec(self):
        return self._environment.reward_spec()

    def close(self):
        return self._environment.close()



from typing import Optional

import numpy as np
from dm_control import mjcf
from dm_control.composer.initializers import utils
from dm_env import specs

from dexterity import exception
from dexterity import goal
from dexterity.models.hands import dexterous_hand
from dexterity.utils import mujoco_collisions
from dexterity.utils import mujoco_utils

_REJECTION_SAMPLING_FAILED = (
    "Failed to find a collision-free initial configuration for the fingertips after "
    "{max_rejection_samples} randomly sampled joint configurations."
)


class FingertipCartesianPosition(goal.GoalGenerator):
    def __init__(
        self,
        hand: dexterous_hand.DexterousHand,
        max_rejection_samples: int = 100,
        scale: float = 0.1,
        name: str = "fingertip_position_goal_generator",
    ) -> None:
        super().__init__()

        self._hand = hand
        self._max_rejection_samples = max_rejection_samples
        self._scale = scale
        self._name = name

        self._qpos: Optional[np.ndarray] = None
        self._reference_qpos: Optional[np.ndarray] = None
        self._goal_spec = None

    def goal_spec(self) -> specs.Array:
        if self._goal_spec is None:
            self._goal_spec = specs.Array(
                shape=(15,), dtype=np.float64, name=self._name
            )
        return self._goal_spec

    def _has_self_collisions(self, physics: mjcf.Physics) -> bool:
        """Returns True if the hand is in a self-collision state."""
        return mujoco_collisions.has_self_collision(physics, self._hand.name)

    def initialize_episode(
        self, physics: mjcf.Physics, random_state: np.random.RandomState
    ) -> None:
        del random_state  # Unused.

        # Apply gravity compensation.
        mujoco_utils.compensate_gravity(physics, self._hand.mjcf_model.find_all("body"))

    def current_state(self, physics: mjcf.Physics) -> np.ndarray:
        return np.array(physics.bind(self._hand.fingertip_sites).xpos).ravel()

    def next_goal(
        self, physics: mjcf.Physics, random_state: np.random.RandomState
    ) -> np.ndarray:
        joint_binding = physics.bind(self._hand.joints)
        actuator_binding = physics.bind(self._hand.actuators)

        # Get joint limits and range.
        joint_limits = joint_binding.range
        joint_range = joint_limits[:, 1] - joint_limits[:, 0]

        # Use the midrange of the joint angles as the reference configuration.
        if self._reference_qpos is None:
            self._reference_qpos = joint_limits.mean(axis=1)

        initial_qpos = joint_binding.qpos.copy()
        initial_ctrl = actuator_binding.ctrl.copy()
        fingertip_pos: Optional[np.ndarray] = None

        for _ in range(self._max_rejection_samples):
            # Sample around the reference configuration in joint space.
            qpos_desired = random_state.normal(
                loc=self._reference_qpos, scale=self._scale * joint_range
            )
            np.clip(
                qpos_desired, joint_limits[:, 0], joint_limits[:, 1], out=qpos_desired
            )

            self._hand.set_joint_angles(physics, qpos_desired)
            physics.forward()

            # Take a few steps to avoid goals that are impossible due to contact.
            original_time = physics.data.time
            ctrl_desired = self._hand.joint_positions_to_control(qpos_desired)
            actuator_binding.ctrl[:] = ctrl_desired
            with utils.JointStaticIsolator(physics, self._hand.joints):
                for _ in range(2):
                    physics.step()

            if not self._has_self_collisions(physics):
                qpos_desired = joint_binding.qpos.copy()
                fingertip_pos = physics.bind(self._hand.fingertip_sites).xpos.copy()
                self._qpos = qpos_desired
                break

            physics.data.time = original_time

        # Restore the initial joint configuration and ctrl.
        actuator_binding.ctrl[:] = initial_ctrl
        self._hand.set_joint_angles(physics, initial_qpos)
        physics.forward()

        if fingertip_pos is None:
            raise exception.GoalInitializationError(
                _REJECTION_SAMPLING_FAILED.format(
                    max_rejection_samples=self._max_rejection_samples
                )
            )
        else:
            return fingertip_pos.ravel()

    def relative_goal(
        self, goal_state: np.ndarray, current_state: np.ndarray
    ) -> np.ndarray:
        return goal_state - current_state

    def goal_distance(
        self, goal_state: np.ndarray, current_state: np.ndarray
    ) -> np.ndarray:
        relative_goal = self.relative_goal(goal_state, current_state).reshape(-1, 3)
        return np.linalg.norm(relative_goal, axis=1)

    @property
    def name(self) -> str:
        return self._name

    @property
    def qpos(self) -> Optional[np.ndarray]:
        """The joint configuration used to place the target sites."""
        return self._qpos



import re
from collections import OrderedDict
from typing import List, Optional, Sequence, Tuple

import numpy as np
from dm_control import composer
from dm_control import mjcf
from dm_env import specs

from dexterity import effector
from dexterity import goal
from dexterity.models.hands import dexterous_hand


class Task(composer.Task):
    """Base class for dexterous manipulation tasks."""

    def __init__(
        self,
        arena: composer.Arena,
        hands: Sequence[dexterous_hand.DexterousHand],
        hand_effectors: Sequence[effector.Effector],
    ) -> None:
        self._arena = arena
        self._hands = tuple(hands)
        self._hand_effectors = tuple(hand_effectors)

        self._action_spec = None

    # Reference: https://github.com/deepmind/dm_robotics/blob/main/py/moma/subtask_env.py
    def _find_effector_indices(
        self, eff: effector.Effector, physics: mjcf.Physics
    ) -> List[bool]:
        action_spec = self.action_spec(physics)
        actuator_names = action_spec.name.split("\t")
        prefix_expr = re.compile(eff.prefix)
        return [re.match(prefix_expr, name) is not None for name in actuator_names]

    # Composer overrides.

    def after_compile(
        self, physics: mjcf.Physics, random_state: np.random.RandomState
    ) -> None:
        del physics, random_state  # Unused.

        for eff in self._hand_effectors:
            eff.after_compile(self.root_entity.mjcf_model.root)

    def initialize_episode(
        self, physics: mjcf.Physics, random_state: np.random.RandomState
    ) -> None:
        for eff in self._hand_effectors:
            eff.initialize_episode(physics, random_state)

    def before_step(
        self,
        physics: mjcf.Physics,
        action: np.ndarray,
        random_state: np.random.RandomState,
    ) -> None:
        del random_state  # Unused.

        for eff in self._hand_effectors:
            e_cmd = action[self._find_effector_indices(eff, physics)]
            eff.set_control(physics, e_cmd)

    def action_spec(self, physics: mjcf.Physics) -> specs.BoundedArray:
        if self._action_spec is None:
            # Merge action specs.
            a_specs = [a.action_spec(physics) for a in self._hand_effectors]
            self._action_spec = merge_specs(a_specs)
        return self._action_spec

    # Accessors.

    @property
    def root_entity(self) -> composer.Entity:
        return self._arena

    @property
    def arena(self) -> composer.Arena:
        return self._arena

    @property
    def hands(self) -> Tuple[dexterous_hand.DexterousHand, ...]:
        return self._hands

    @property
    def hand_effectors(self) -> Tuple[effector.Effector, ...]:
        return self._hand_effectors

    @property
    def step_limit(self) -> Optional[int]:
        """The maximum number of steps in an episode."""
        return None

    @property
    def time_limit(self) -> float:
        """The maximum number of seconds in an episode."""
        return float("inf")


class GoalTask(Task):
    """Goal reaching based tasks."""

    def __init__(
        self,
        arena: composer.Arena,
        hands: Sequence[dexterous_hand.DexterousHand],
        hand_effectors: Sequence[effector.Effector],
        goal_generator: goal.GoalGenerator,
        success_threshold: float,
        successes_needed: int = 1,
        steps_before_changing_goal: int = 0,
        max_time_per_goal: Optional[float] = None,
    ) -> None:
        super().__init__(arena, hands, hand_effectors)

        self._goal_generator = goal_generator
        self._steps_before_changing_goal = steps_before_changing_goal
        self._successes_needed = successes_needed
        self._max_time_per_goal = max_time_per_goal
        self._success_threshold = success_threshold

        # Initialize with dummy goal to appease `task_observables`.
        self._goal = self._goal_generator.goal_spec().generate_value()

    def initialize_episode(
        self, physics: mjcf.Physics, random_state: np.random.RandomState
    ) -> None:
        super().initialize_episode(physics, random_state)

        self._goal_generator.initialize_episode(physics, random_state)

        # Generate the first goal.
        self._goal = self._goal_generator.next_goal(physics, random_state)

        self._successes = 0
        self._success_change_counter = 0
        self._solve_start_time = physics.data.time
        self._exceeded_single_goal_time = False
        self._success_registered = False
        self._goal_changed = True

    def before_step(self, physics, action, random_state):
        super().before_step(physics, action, random_state)

        if self._success_change_counter > self._steps_before_changing_goal:
            self._goal = self._goal_generator.next_goal(physics, random_state)
            self._success_change_counter = 0
            self._exceeded_single_goal_time = False
            self._solve_start_time = physics.data.time
            self._goal_changed = True
            self._success_registered = False
        else:
            self._goal_changed = False

    def after_step(
        self, physics: mjcf.Physics, random_state: np.random.RandomState
    ) -> None:
        del random_state  # Unused.

        self._goal_distance = self._goal_generator.goal_distance(
            self._goal,
            self._goal_generator.current_state(physics),
        )

        if np.all(self._goal_distance <= self._success_threshold):
            self._success_change_counter += 1
            if not self._success_registered:
                self._successes += 1
                self._success_registered = True
        else:
            if self._max_time_per_goal is not None:
                if physics.data.time - self._solve_start_time > self._max_time_per_goal:
                    self._exceeded_single_goal_time = True

    def should_terminate_episode(self, physics):
        del physics  # Unused.
        if self._successes >= self._successes_needed or (
            self._max_time_per_goal is not None and self._exceeded_single_goal_time
        ):
            return True
        return False

    def get_discount(self, physics: mjcf.Physics) -> float:
        # In the finite-horizon setting, on successful termination, we return 0.0 to
        # indicate a terminal state. If the episode did not successfully terminate,
        # i.e., the agent exceeded the time limit for a single solve, we return a
        # discount of 1.0 to indicate that the agent should treat the episode as if it
        # would have continued, even though the trajectory is truncated.
        del physics  # Unused.
        if self._successes >= self._successes_needed:
            return 0.0
        return 1.0

    @property
    def task_observables(self) -> OrderedDict:
        task_observables = OrderedDict()

        # Add the goal at the current timestep to the task observables.
        goal_spec = self._goal_generator.goal_spec()
        goal_observable = goal.GoalObservable(lambda _: np.array(self._goal), goal_spec)
        goal_observable.enabled = True
        task_observables["goal_state"] = goal_observable

        return task_observables

    @property
    def goal_generator(self) -> goal.GoalGenerator:
        return self._goal_generator

    @property
    def successes(self) -> int:
        return self._successes

    @property
    def successes_needed(self) -> int:
        return self._successes_needed


# Reference: https://github.com/deepmind/dm_robotics/blob/main/py/agentflow/spec_utils.py
def merge_specs(spec_list: Sequence[specs.BoundedArray]):
    """Merges a list of `BoundedArray` specs into one."""

    # Check all specs are flat.
    for spec in spec_list:
        if len(spec.shape) > 1:
            raise ValueError("Not merging multi-dimensional spec: {}".format(spec))

    # Filter out no-op specs with no actuators.
    spec_list = [spec for spec in spec_list if spec.shape and spec.shape[0]]
    dtype = np.find_common_type([spec.dtype for spec in spec_list], [])

    num_actions = 0
    name = ""
    mins = np.array([], dtype=dtype)
    maxs = np.array([], dtype=dtype)

    for i, spec in enumerate(spec_list):
        num_actions += spec.shape[0]
        if name:
            name += "\t"
        name += spec.name or f"spec_{i}"
        mins = np.concatenate([mins, spec.minimum])
        maxs = np.concatenate([maxs, spec.maximum])

    return specs.BoundedArray(
        shape=(num_actions,), dtype=dtype, minimum=mins, maximum=maxs, name=name
    )



from pathlib import Path

__version__ = "0.0.7"

# Path to the root of the project.
_PROJECT_ROOT: Path = Path(__file__).parent.parent

# Path to the root of the src files, i.e. `dexterity/`.
_SRC_ROOT: Path = _PROJECT_ROOT / "dexterity"



"""Tests for reach.py."""

import numpy as np
from absl.testing import absltest

from dexterity.environment import GoalEnvironment
from dexterity.manipulation.shared import observations
from dexterity.manipulation.tasks import reach_task


class ReachTaskTest(absltest.TestCase):
    def test_sparse_reward(self) -> None:
        task = reach_task(
            observations.ObservationSet.STATE_ONLY,
            use_dense_reward=False,
            visualize_reward=False,
        )

        random_state = np.random.RandomState(12345)
        env = GoalEnvironment(task, random_state=random_state)
        action_spec = env.action_spec()

        env.reset()
        timestep = env.step(np.zeros(action_spec.shape, action_spec.dtype))
        self.assertEqual(timestep.reward, -1.0)

        qpos_sol = env.task.goal_generator.qpos.copy()  # type: ignore
        ctrl_sol = env.task.hand.joint_positions_to_control(qpos_sol)
        ctrl_sol = ctrl_sol.astype(action_spec.dtype)
        while True:
            timestep = env.step(ctrl_sol)
            if env.task.successes > 0:
                break
        expected_reward = 0.0
        np.testing.assert_equal(timestep.reward, expected_reward)


if __name__ == "__main__":
    absltest.main()



"""Common type hints for dexterity."""

import mujoco
from dm_control import mjcf
from dm_control.mujoco.wrapper.core import MjData
from dm_control.mujoco.wrapper.core import MjModel
from typing_extensions import TypeAlias

MjcfElement: TypeAlias = mjcf.element._ElementImpl
MjcfAttachmentFrame: TypeAlias = mjcf.element._AttachmentFrame
MujocoModel: TypeAlias = MjModel
MujocoData: TypeAlias = MjData
MujocoObjectType: TypeAlias = mujoco.mjtObj



"""Tests for JugglingBall."""

from absl.testing import absltest
from dm_control import mjcf

from dexterity.manipulation.props import JugglingBall


class JugglingBallTest(absltest.TestCase):
    def test_can_compile_and_step_model(self) -> None:
        prop = JugglingBall()
        physics = mjcf.Physics.from_mjcf_model(prop.mjcf_model)
        physics.step()


if __name__ == "__main__":
    absltest.main()



"""Tests for geometry_utils."""

import numpy as np
from absl.testing import absltest

from dexterity.utils import geometry_utils


class GeometryUtilsTest(absltest.TestCase):
    def test_l2_normalize(self) -> None:
        input_array = np.random.randn(4)
        output_array = geometry_utils.l2_normalize(input_array)
        self.assertTrue(np.isclose(np.linalg.norm(output_array), 1.0))


if __name__ == "__main__":
    absltest.main()



"""Tests for ik_solver."""

import numpy as np
from absl.testing import absltest

from dexterity.inverse_kinematics import ik_solver
from dexterity.models.hands import shadow_hand_e
from dexterity.models.hands import shadow_hand_e_constants as consts


class IKSolverTest(absltest.TestCase):
    def test_return_none_when_passing_impossible_target(self) -> None:
        hand = shadow_hand_e.ShadowHandSeriesE()
        target_positions = {
            consts.Components.TH: np.array([5.0, 5.0, 5.0]),
        }
        solver = ik_solver.IKSolver(
            hand.mjcf_model, fingers=tuple(target_positions.keys())
        )
        qpos_sol = solver.solve(target_positions)
        self.assertIsNone(qpos_sol)


if __name__ == "__main__":
    absltest.main()



from dexterity.manipulation.goals.fingertip_position import FingertipCartesianPosition
from dexterity.manipulation.goals.prop_orientation import PropOrientation

__all__ = [
    "FingertipCartesianPosition",
    "PropOrientation",
]



import abc
from typing import Callable

import numpy as np
from dm_control import mjcf
from dm_control.composer.observation.observable import Observable
from dm_env import specs


class GoalObservable(Observable):
    def __init__(
        self,
        raw_observation_callable: Callable[[mjcf.Physics], np.ndarray],
        goal_spec: specs.Array,
        update_interval=1,
        buffer_size=None,
        delay=None,
        aggregator=None,
        corruptor=None,
    ) -> None:

        self._raw_callable = raw_observation_callable
        self._goal_spec = goal_spec

        super().__init__(update_interval, buffer_size, delay, aggregator, corruptor)

    @property
    def array_spec(self) -> specs.Array:
        return self._goal_spec

    def _callable(self, physics: mjcf.Physics) -> Callable[[], np.ndarray]:
        return lambda: self._raw_callable(physics)


class GoalGenerator(abc.ABC):
    """Abstract base class for a goal generator."""

    @abc.abstractmethod
    def goal_spec(self) -> specs.Array:
        ...

    @abc.abstractmethod
    def initialize_episode(
        self, physics: mjcf.Physics, random_state: np.random.RandomState
    ) -> None:
        ...

    @abc.abstractmethod
    def next_goal(
        self, physics: mjcf.Physics, random_state: np.random.RandomState
    ) -> np.ndarray:
        ...

    @abc.abstractmethod
    def relative_goal(
        self, goal_state: np.ndarray, current_state: np.ndarray
    ) -> np.ndarray:
        ...

    @abc.abstractmethod
    def goal_distance(
        self, goal_state: np.ndarray, current_state: np.ndarray
    ) -> np.ndarray:
        ...

    @abc.abstractmethod
    def current_state(self, physics: mjcf.Physics) -> np.ndarray:
        ...

    @property
    @abc.abstractmethod
    def name(self) -> str:
        ...



from typing import Optional, Sequence

from dm_control import composer

from dexterity import hints


class Arena(composer.Arena):
    """Standard composer arena with added functionality."""

    def attach_offset(
        self,
        entity: composer.Entity,
        position: Optional[Sequence[float]] = None,
        quaternion: Optional[Sequence[float]] = None,
        attach_site: Optional[hints.MjcfAttachmentFrame] = None,
    ):
        frame = self.attach(entity, attach_site=attach_site)
        if position is not None:
            if len(position) != 3:
                raise ValueError("Position must be a sequence of length 3.")
            frame.pos = position
        if quaternion is not None:
            if len(quaternion) != 4:
                raise ValueError("Quaternion must be a sequence of length 4.")
            frame.quat = quaternion
        return frame

    def add_mocap(
        self,
        entity: composer.Entity,
        position: Optional[Sequence[float]] = None,
        quaternion: Optional[Sequence[float]] = None,
        visible: bool = False,
    ) -> None:
        # Add mocap body.
        mocap_body = self.mjcf_model.worldbody.add(
            "body",
            name="mocap",
            mocap="true",
            pos=position,
            quat=quaternion,
        )

        # Give it a square shape.
        if visible:
            mocap_body.add(
                "geom",
                type="box",
                group="1",
                size="0.02 0.02 0.02",
                contype="0",
                conaffinity="0",
                rgba=".9 .5 .5 1",
            )

        # Make root body of entity's pose same as mocap pose.
        root_body = entity.mjcf_model.find_all("body")[0]
        root_body.pos = position
        root_body.quat = quaternion

        # Add the entity to the arena as a free body.
        self.add_free_entity(entity)

        # Add a weld constraint between the mocap body and the root body.
        self.mjcf_model.equality.add(
            "weld",
            body1="mocap",
            body2=root_body.full_identifier,
            solref="0.01 1",
            solimp=".9 .9 0.01",
        )



"""A standalone app for visualizing manipulation tasks."""

from typing import Sequence

from absl import app
from absl import flags
from dm_control import viewer

from dexterity import manipulation
from dexterity.manipulation.wrappers import ActionNoise

flags.DEFINE_enum(
    "environment_name",
    None,
    manipulation.ALL_NAMES,
    "Optional 'domain_name.task_name' pair specifying the environment to load.",
)
flags.DEFINE_integer("seed", None, "RNG seed.")
flags.DEFINE_bool("timeout", True, "Whether episodes should have a time limit.")
flags.DEFINE_float("action_noise", 0.0, "Action noise scale.")

FLAGS = flags.FLAGS


def prompt_environment_name(values: Sequence[str]) -> str:
    environment_name = None
    while not environment_name:
        environment_name = input("Please enter the environment name: ")
        if not environment_name or environment_name not in values:
            print(f"'{environment_name}' is not a valid environment name.")
            environment_name = None
    return environment_name


def main(_) -> None:
    if FLAGS.environment_name is None:
        print("\n ".join(["Available environments:"] + manipulation.ALL_NAMES))
        environment_name = prompt_environment_name(manipulation.ALL_NAMES)
    else:
        environment_name = FLAGS.environment_name

    index = manipulation.ALL_NAMES.index(environment_name)
    domain_name, task_name = manipulation.ALL_TASKS[index]

    env = manipulation.load(
        domain_name=domain_name,
        task_name=task_name,
        seed=FLAGS.seed,
        time_limit=float("inf") if not FLAGS.timeout else None,
    )

    if FLAGS.action_noise > 0.0:
        # By default, the viewer will apply the midpoint action since the action spec is
        # bounded. So this wrapper will add noise centered around this midpoint.
        env = ActionNoise(env, scale=FLAGS.action_noise)

    # Print entity and task observables.
    timestep = env.reset()
    for k, v in timestep.observation.items():
        print(f"{k}: {v.shape}")

    viewer.launch(env)


if __name__ == "__main__":
    app.run(main)



from dexterity.manipulation.tasks.juggle import Juggle
from dexterity.manipulation.tasks.juggle import juggle_task
from dexterity.manipulation.tasks.reach import Reach
from dexterity.manipulation.tasks.reach import reach_task
from dexterity.manipulation.tasks.reorient import ReOrient
from dexterity.manipulation.tasks.reorient import reorient_task

__all__ = [
    "Reach",
    "reach_task",
    "ReOrient",
    "reorient_task",
    "Juggle",
    "juggle_task",
]



"""Tests for the standard."""

from absl.testing import absltest
from dm_control import mjcf

from dexterity.manipulation.arenas import standard


class StandardTest(absltest.TestCase):
    def test_can_compile_and_step_model(self) -> None:
        arena = standard.Standard()
        physics = mjcf.Physics.from_mjcf_model(arena.mjcf_model)
        physics.step()


if __name__ == "__main__":
    absltest.main()



import fnmatch
import os
import re
from pathlib import Path

from setuptools import find_packages
from setuptools import setup

_here = Path(__file__).resolve().parent

name = "dexterity"

# Reference: https://github.com/patrick-kidger/equinox/blob/main/setup.py
with open(_here / name / "__init__.py") as f:
    meta_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", f.read(), re.M)
    if meta_match:
        version = meta_match.group(1)
    else:
        raise RuntimeError("Unable to find __version__ string.")


with open(_here / "README.md", "r") as f:
    readme = f.read()

core_requirements = [
    "absl-py",
    "numpy",
    "typing_extensions",
    "mujoco",
    "dm_control >= 1.0.1",
    "dm_robotics-geometry",
    "dm_robotics-transformations",
]

examples_requirements = [
    "matplotlib",
    "imageio",
    "imageio-ffmpeg",
]

testing_requirements = [
    "pytest-xdist",
]

dev_requirements = (
    [
        "black",
        "isort",
        "flake8",
        "mypy",
        "ipdb",
        "jupyter",
    ]
    + testing_requirements
    + examples_requirements
)

classifiers = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Developers",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: MIT License",
    "Natural Language :: English",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Topic :: Scientific/Engineering :: Artificial Intelligence",
]

author = "Kevin Zakka"

author_email = "kevinarmandzakka@gmail.com"

description = "Software and tasks for dexterous multi-fingered hand manipulation, powered by MuJoCo"


# Reference: https://github.com/deepmind/dm_control/blob/main/setup.py
def find_data_files(package_dir, patterns, excludes=()):
    """Recursively finds files whose names match the given shell patterns."""
    paths = set()

    def is_excluded(s):
        for exclude in excludes:
            if fnmatch.fnmatch(s, exclude):
                return True
        return False

    for directory, _, filenames in os.walk(package_dir):
        if is_excluded(directory):
            continue
        for pattern in patterns:
            for filename in fnmatch.filter(filenames, pattern):
                # NB: paths must be relative to the package directory.
                relative_dirpath = os.path.relpath(directory, package_dir)
                full_path = os.path.join(relative_dirpath, filename)
                if not is_excluded(full_path):
                    paths.add(full_path)

    return list(paths)


setup(
    name=name,
    version=version,
    author=author,
    author_email=author_email,
    maintainer=author,
    maintainer_email=author_email,
    description=description,
    long_description=readme,
    long_description_content_type="text/markdown",
    url=f"https://github.com/kevinzakka/{name}",
    license="MIT",
    license_files=("LICENSE",),
    packages=find_packages(),
    package_data={
        "dexterity": find_data_files(
            package_dir="dexterity",
            patterns=["*.msh", "*.png", "*.skn", "*.stl", "*.xml", "*.typed"],
            excludes=[],
        ),
    },
    zip_safe=True,
    python_requires=">=3.7",
    install_requires=core_requirements,
    classifiers=classifiers,
    extras_require={
        "testing": testing_requirements,
        "examples": examples_requirements,
        "dev": dev_requirements,
    },
)



"""Shared configuration options for """

import dataclasses
import enum
from typing import Callable, Optional, Tuple, Union


@dataclasses.dataclass(frozen=True)
class ObservableSpec:
    """Configuration options for generic observables."""

    enabled: bool
    update_interval: Union[int, Callable[..., int]]
    buffer_size: int
    delay: Union[int, Callable[..., int]]
    aggregator: Optional[Union[str, Callable[..., int]]]
    corruptor: Optional[Callable[..., int]]


@dataclasses.dataclass(frozen=True)
class CameraObservableSpec(ObservableSpec):
    """Configuration options for camera observables."""

    height: int
    width: int
    depth: bool
    segmentation: bool


@dataclasses.dataclass(frozen=True)
class ObservationSettings:
    """Container for `ObservableSpec`s grouped by category."""

    privileged_proprio: ObservableSpec
    proprio: ObservableSpec
    prop_pose: ObservableSpec
    camera: CameraObservableSpec


@dataclasses.dataclass(frozen=True)
class ObservableNames:
    """Container that groups the names of observables by category."""

    privileged_proprio: Tuple[str, ...] = dataclasses.field(default_factory=tuple)
    proprio: Tuple[str, ...] = dataclasses.field(default_factory=tuple)
    prop_pose: Tuple[str, ...] = dataclasses.field(default_factory=tuple)
    camera: Tuple[str, ...] = dataclasses.field(default_factory=tuple)


# Global defaults for feature observables (i.e., anything that isn't a camera).
_DISABLED_FEATURE = ObservableSpec(
    enabled=False,
    update_interval=1,
    buffer_size=1,
    delay=0,
    aggregator=None,
    corruptor=None,
)
_ENABLED_FEATURE = dataclasses.replace(_DISABLED_FEATURE, enabled=True)

# Global defaults for camera observables.
_DISABLED_CAMERA = CameraObservableSpec(
    height=84,
    width=84,
    depth=False,
    segmentation=False,
    enabled=False,
    update_interval=1,
    buffer_size=1,
    delay=0,
    aggregator=None,
    corruptor=None,
)
_ENABLED_CAMERA = dataclasses.replace(_DISABLED_CAMERA, enabled=True)

# Predefined sets of configurations to apply to each category of observable.
_STATE_ONLY = ObservationSettings(
    privileged_proprio=_ENABLED_FEATURE,
    proprio=_ENABLED_FEATURE,
    prop_pose=_ENABLED_FEATURE,
    camera=_DISABLED_CAMERA,
)
_VISION_ONLY = ObservationSettings(
    privileged_proprio=_DISABLED_FEATURE,
    proprio=_ENABLED_FEATURE,
    prop_pose=_DISABLED_FEATURE,
    camera=_ENABLED_CAMERA,
)
_ALL = ObservationSettings(
    privileged_proprio=_ENABLED_FEATURE,
    proprio=_ENABLED_FEATURE,
    prop_pose=_ENABLED_FEATURE,
    camera=_ENABLED_CAMERA,
)

HAND_OBSERVABLES = ObservableNames(
    privileged_proprio=(
        "joint_velocities",
        "fingertip_positions",
        "fingertip_linear_velocities",
    ),
    proprio=("joint_positions_sin_cos",),
)


class ObservationSet(enum.Enum):
    """Different possible set of observations that can be exposed."""

    STATE_ONLY = _STATE_ONLY
    VISION_ONLY = _VISION_ONLY
    ALL = _ALL


def make_options(obs_settings: ObservationSettings, obs_names: ObservableNames):
    """Constructs a dict of configuration options for a set of named observables."""
    observable_options = {}
    for category, spec in dataclasses.asdict(obs_settings).items():
        for observable_name in getattr(obs_names, category):
            observable_options[observable_name] = spec
    return observable_options



import dataclasses
from typing import Mapping

import numpy as np


@dataclasses.dataclass(frozen=True)
class Reward:
    value: float
    weight: float


def weighted_average(rewards: Mapping[str, Reward]) -> float:
    """Computes a weighted average of shaped reward components."""
    return sum([reward.value * reward.weight for reward in rewards.values()])


def tanh_squared(x: np.ndarray, margin: float, loss_at_margin: float = 0.95):
    if not margin > 0:
        raise ValueError("`margin` must be positive.")
    if not 0.0 < loss_at_margin < 1.0:
        raise ValueError("`loss_at_margin` must be between 0 and 1.")

    error = np.linalg.norm(x)
    # Compute weight such that at the margin tanh(w * error) = loss_at_margin
    w = np.arctanh(np.sqrt(loss_at_margin)) / margin
    s = np.tanh(w * error)
    return s * s



from typing import List, Sequence

import mujoco
import numpy as np
from dm_control import mjcf
from dm_robotics.transformations import transformations as tr

from dexterity import hints


def prefix_identifier(identifier: str, prefix: str) -> str:
    """Prefixes the identifier with the provided string.

    This helper function deals with possibly empty prefixes and does not add the slash
    delimiter.
    """
    if not prefix:
        return identifier
    return f"{prefix}/{identifier}"


def get_site_pose(physics: mjcf.Physics, site_elem: hints.MjcfElement) -> np.ndarray:
    """Returns the world pose of the site as a 4x4 transform.

    Args:
        physics: An `mjcf.Physics` instance.
        site_entity: An `mjcf.Element` instance.
    """
    binding = physics.bind(site_elem)
    xpos = binding.xpos.reshape(3, 1)
    xmat = binding.xmat.reshape(3, 3)
    return np.vstack(
        [
            np.hstack([xmat, xpos]),
            np.array([0, 0, 0, 1]),
        ]
    )


def get_site_relative_pose(
    physics: mjcf.Physics, site_a: hints.MjcfElement, site_b: hints.MjcfElement
) -> np.ndarray:
    """Returns the pose of `site_a` in the frame of `site_b`.

    Args:
        physics: An `mjcf.Physics` instance.
        site_a: An `mjcf.Element` instance.
        site_b: An `mjcf.Element` instance.
    """
    pose_wa = get_site_pose(physics, site_a)  # Pose of site_a in world frame.
    pose_wb = get_site_pose(physics, site_b)  # Pose of site_b in world frame.
    pose_bw = tr.hmat_inv(pose_wb)  # Pose of world frame in site_b.
    pose_ba = pose_bw @ pose_wa  # Pose of site_a in site_b.
    return pose_ba


def get_site_velocity(
    physics: mjcf.Physics, site_elem: mjcf.Element, world_frame: bool = False
) -> np.ndarray:
    """Returns the linear and angular velocities of the site.

    This 6-D vector represents the instantaneous velocity of the coordinate system
    attached to the site. If `world_frame=True`, the velocity is expressed in the world
    frame.

    Args:
        physics: An `mjcf.Physics` instance.
        site_elem: An `mjcf.Element` instance.
        world_frame: Whether to return the velocity in the world frame.
    """
    flg_local = 0 if world_frame else 1
    idx = physics.model.name2id(site_elem.full_identifier, mujoco.mjtObj.mjOBJ_SITE)
    site_vel = np.empty(6)
    mujoco.mj_objectVelocity(
        physics.model.ptr,
        physics.data.ptr,
        mujoco.mjtObj.mjOBJ_SITE,
        idx,
        site_vel,
        flg_local,
    )
    return np.hstack([site_vel[3:], site_vel[:3]])


def get_joint_dof_size(model: hints.MjModel, joint_id: int) -> int:
    """Returns the number of degrees of freedom of the joint."""
    if joint_id < 0 or joint_id >= model.njnt:
        raise ValueError(
            f"Provided joint_id {joint_id} is invalid for the provided model, "
            f"which has {model.njnt} joints."
        )

    joint_type = model.jnt_type[joint_id]

    if joint_type == mujoco.mjtJoint.mjJNT_SLIDE:
        return 1
    elif joint_type == mujoco.mjtJoint.mjJNT_HINGE:
        return 1
    elif joint_type == mujoco.mjtJoint.mjJNT_BALL:
        return 3
    elif joint_type == mujoco.mjtJoint.mjJNT_FREE:
        return 6
    else:  # Not supported.
        return -1


def joint_ids_to_dof_ids(model: hints.MjModel, joint_ids: Sequence[int]) -> List[int]:
    dof_ids = []
    for joint_id in joint_ids:
        dof = get_joint_dof_size(model, joint_id)
        if dof == -1:
            raise ValueError(f"joint_id {joint_id} is not a recognized joint type.")
        for i in range(dof):
            dof_ids.append(model.jnt_dofadr[joint_id] + i)
    return dof_ids


def compute_object_6d_jacobian(
    model: hints.MjModel,
    data: hints.MjData,
    object_type: hints.MujocoObjectType,
    object_id: int,
) -> np.ndarray:
    """Computes the (6, nv) object Jacobian.

    The Jacobian maps joint velocities to the object's Cartesian 6D velocity in the
    world frame. Each row of the Jacobian is the gradient of the corresponding 3D
    coordinate of the specified point with respect to the degrees of freedom.

    Only MuJoCo bodies, geoms and sites are supported.
    """
    jacobian = np.empty((6, model.nv), dtype=data.qpos.dtype)
    jacobian_position, jacobian_rotation = jacobian[:3], jacobian[3:]

    if object_type == mujoco.mjtObj.mjOBJ_BODY:
        func = mujoco.mj_jacBody
    elif object_type == mujoco.mjtObj.mjOBJ_GEOM:
        func = mujoco.mj_jacGeom
    elif object_type == mujoco.mjtObj.mjOBJ_SITE:
        func = mujoco.mj_jacSite
    else:
        raise ValueError(
            f"Invalid `object_type` {object_type}. Only bodies, geoms and sites"
            " are supported."
        )

    func(
        model.ptr,
        data.ptr,
        jacobian_position,
        jacobian_rotation,
        object_id,
    )

    return jacobian


def get_element_type(element: hints.MjcfElement) -> mujoco.mjtObj:
    if element.tag == "body":
        return mujoco.mjtObj.mjOBJ_BODY
    elif element.tag == "geom":
        return mujoco.mjtObj.mjOBJ_GEOM
    elif element.tag == "site":
        return mujoco.mjtObj.mjOBJ_SITE
    else:
        raise ValueError(
            f"Element must be a MuJoCo body, geom or site. Got [{element.tag}]."
        )


def compensate_gravity(
    physics: mjcf.Physics, body_elements: Sequence[hints.MjcfElement]
) -> None:
    """Counteracts gravity by applying forces to body elements."""
    gravity = np.hstack([physics.model.opt.gravity, [0.0, 0.0, 0.0]])
    physics_bodies = physics.bind(body_elements)
    if physics_bodies is None:
        raise ValueError("Calling bind() on the body elements returned None.")
    physics_bodies.xfrc_applied[:] = -gravity * physics_bodies.mass[..., None]



from typing import List

import numpy as np
from dm_control import composer
from dm_control import mjcf

from dexterity.hints import MjcfElement
from dexterity.models.hands import adroit_hand_constants as consts
from dexterity.models.hands import dexterous_hand
from dexterity.utils import mujoco_utils


class AdroitHand(dexterous_hand.DexterousHand):
    def _build(
        self,
        name: str = "adroit_hand",
    ) -> None:
        self._mjcf_root = mjcf.from_path(str(consts.ADROIT_HAND_E_XML))
        self._mjcf_root.model = name

        self._parse_mjcf_elements()

    def initialize_episode(
        self, physics: mjcf.Physics, random_state: np.random.RandomState
    ) -> None:
        del random_state  # Unused.

        # Apply gravity compensation.
        mujoco_utils.compensate_gravity(physics, self.mjcf_model.find_all("body"))

    # ================= #
    # Accessors.
    # ================= #

    @property
    def mjcf_model(self) -> mjcf.RootElement:
        return self._mjcf_root

    @property
    def name(self) -> str:
        return self._mjcf_root.model

    @property
    def joints(self) -> List[MjcfElement]:
        """List of joint elements belonging to the hand."""
        return self._joints

    @property
    def actuators(self) -> List[MjcfElement]:
        """List of actuator elements belonging to the hand."""
        return self._actuators

    @property
    def tendons(self) -> List[MjcfElement]:
        """List of tendon elements belonging to the hand."""
        return self._tendons

    @property
    def joint_torque_sensors(self) -> List[MjcfElement]:
        """List of joint torque sensor elements belonging to the hand."""
        return self._joint_torque_sensors

    @property
    def fingertip_sites(self) -> List[MjcfElement]:
        """List of fingertip site elements belonging to the hand."""
        return self._fingertip_sites

    # ================= #
    # Public methods.
    # ================= #

    def control_to_joint_positions(self, control: np.ndarray) -> np.ndarray:
        # The Adroit hand is fully-actuated, so ctrl = qpos.
        return control

    def joint_positions_to_control(self, qpos: np.ndarray) -> np.ndarray:
        # The Adroit hand is fully-actuated, so ctrl = qpos.
        return qpos

    def set_joint_angles(self, physics: mjcf.Physics, joint_angles: np.ndarray) -> None:
        physics.bind(self._joints).qpos = joint_angles

    def sample_joint_angles(
        self, physics: mjcf.Physics, random_state: np.random.RandomState
    ) -> np.ndarray:
        return random_state.uniform(*physics.bind(self._joints).range.T)

    # ================= #
    # Private methods.
    # ================= #

    def _parse_mjcf_elements(self) -> None:
        """Parses MJCF elements that will be exposed as attributes."""

        self._joints: List[mjcf.Element] = self._mjcf_root.find_all("joint")
        if not self._joints:
            raise ValueError("No joints found in the MJCF model.")

        self._tendons: List[mjcf.Element] = self._mjcf_root.find_all("tendon")
        if not self._tendons:
            raise ValueError("No tendons found in the MJCF model.")

        self._actuators: List[mjcf.Element] = self._mjcf_root.find_all("actuator")
        if not self._actuators:
            raise ValueError("No actuators found in the MJCF model.")

        self._fingertip_sites: List[mjcf.Element] = []
        for fingertip_site_name in consts.FINGERTIP_SITE_NAMES:
            fingertip_site_elem = self._mjcf_root.find("site", fingertip_site_name)
            if fingertip_site_elem is None:
                raise ValueError(
                    f"No fingertip site found with name {fingertip_site_name}."
                )
            self._fingertip_sites.append(fingertip_site_elem)

        self._joint_torque_sensors: List[mjcf.Element] = []
        for joint_elem in self._joints:
            site_elem = joint_elem.parent.add(
                "site",
                name=joint_elem.name + "_site",
                # NOTE(kevin): The kwargs below are for visualization purposes.
                size="0.001 0.001 0.001",
                type="box",
                rgba="0 1 0 1",
                group=composer.SENSOR_SITES_GROUP,
            )
            # Create a 3-axis torque sensor.
            torque_sensor_elem = joint_elem.root.sensor.add(
                "torque",
                site=site_elem,
                name=joint_elem.name + "_torque",
            )
            self._joint_torque_sensors.append(torque_sensor_elem)



import abc

import numpy as np
from dm_control import mjcf
from dm_env import specs


class Effector(abc.ABC):
    """Abstract base class for an effector."""

    @abc.abstractmethod
    def after_compile(self, mjcf_model: mjcf.RootElement) -> None:
        ...

    @abc.abstractmethod
    def initialize_episode(
        self,
        physics: mjcf.Physics,
        random_state: np.random.RandomState,
    ) -> None:
        ...

    @abc.abstractmethod
    def action_spec(self, physics: mjcf.Physics) -> specs.BoundedArray:
        ...

    @abc.abstractmethod
    def set_control(self, physics: mjcf.Physics, command: np.ndarray) -> None:
        ...

    @property
    @abc.abstractmethod
    def prefix(self) -> str:
        ...



from typing import Sequence

import numpy as np
from dm_control import mjcf
from dm_env import specs

from dexterity import effector
from dexterity import hints


class MujocoEffector(effector.Effector):
    """A generic effector for multiple MuJoCo actuators."""

    def __init__(
        self,
        actuators: Sequence[hints.MjcfElement],
        prefix: str = "",
    ):
        self._actuators = actuators
        self._prefix = prefix
        self._action_spec = None

    def action_spec(self, physics: mjcf.Physics) -> specs.BoundedArray:
        if self._action_spec is None:
            self._action_spec = create_action_spec(
                physics, self._actuators, self._prefix
            )
        return self._action_spec

    def set_control(self, physics: mjcf.Physics, command: np.ndarray) -> None:
        # Note: MuJoCo automatically clips out-of-bounds control inputs so we don't
        # need to do anything special here.
        physics.bind(self._actuators).ctrl[:] = command

    @property
    def prefix(self) -> str:
        return self._prefix

    def after_compile(self, mjcf_model: mjcf.RootElement) -> None:
        del mjcf_model  # Unused.

    def initialize_episode(
        self, physics: mjcf.Physics, random_state: np.random.RandomState
    ) -> None:
        del physics, random_state  # Unused.


def create_action_spec(
    physics: mjcf.Physics,
    actuators: Sequence[hints.MjcfElement],
    prefix: str = "",
) -> specs.BoundedArray:
    """Creates an action range for the given actuators."""
    num_actuators = len(actuators)
    actuator_names = [f"{prefix}{i}" for i in range(num_actuators)]
    control_range = physics.bind(actuators).ctrlrange
    is_limited = physics.bind(actuators).ctrllimited.astype(bool)
    action_min = np.full(num_actuators, fill_value=-np.inf, dtype=np.float32)
    action_max = np.full(num_actuators, fill_value=np.inf, dtype=np.float32)
    action_min[is_limited], action_max[is_limited] = control_range[is_limited].T
    return specs.BoundedArray(
        shape=(num_actuators,),
        dtype=np.float32,
        minimum=action_min,
        maximum=action_max,
        name="\t".join(actuator_names),
    )



"""Utility functions for dealing with MuJoCo collisions.

Adapted from:
- [1]: https://github.com/deepmind/dm_robotics/blob/main/py/moma/utils/mujoco_collisions.py
- [2]: https://github.com/deepmind/rgb_stacking/blob/main/rgb_stacking/physics_utils.py
"""

import itertools
from typing import Optional, Sequence, Union

from dm_control import mjcf
from dm_control import mujoco

_DEFAULT_COLLISION_MARGIN: float = 1e-8


def exclude_bodies_based_on_contype_conaffinity(
    mjcf_model: mjcf.RootElement, exclude_exception_str: Optional[str] = None
):
    """Adds a contact-exclude MJCF element for the body pairs described below.

    A contact-exclude MJCF element is added for every body pair for which all of
    their geoms do not pass the contype/conaffinity check. This ensures that the
    contype/conaffinity filtering happens right after the broad-phase collision
    detection mechanism, usually decreasing the amount of geoms that are checked
    through MuJoCo's bounding sphere test. This may improve computation
    significantly in cases where the broad-phase collision detection is
    ineffective in pruning geoms.

    This should be called after mjcf_model has been finalized, to ensure that the
    body names in the exclude element match those of the compiled model.

    Args:
        mjcf_model: mjcf.RootElement of the finalized MuJoCo XML for the scene.
        exclude_exception_str: if this string is found in the name of the body that
            body is exempt from the contact.add exclude operation.
    """
    # We compile the model first to make sure all the contypes/conaffinities are
    # set.
    xml_string = mjcf_model.to_xml_string()
    assets = mjcf_model.get_assets()
    model = mujoco.wrapper.MjModel.from_xml_string(xml_string, assets=assets)

    # We loop through all body pairs, and exclude them if none of the geoms in the
    # first body pass the contype/conaffinity check against the geoms in the
    # second body.
    for body1_id in range(model.nbody):
        body1_name = model.id2name(body1_id, "body")
        if exclude_exception_str is None or exclude_exception_str not in body1_name:
            for body2_id in range(body1_id):
                body2_name = model.id2name(body2_id, "body")
                if (
                    exclude_exception_str is None
                    or exclude_exception_str not in body2_name
                ):
                    if not _is_any_geom_pass_contype_conaffinity_check(
                        model, body1_id, body2_id
                    ):
                        mjcf_model.contact.add(
                            "exclude", body1=body1_name, body2=body2_name
                        )


def _is_any_geom_pass_contype_conaffinity_check(
    model: mujoco.wrapper.MjModel, body1_id: int, body2_id: int
):
    """Returns true if any geom pair passes the contype/conaff check."""
    body1_geomadr = model.body_geomadr[body1_id]
    body2_geomadr = model.body_geomadr[body2_id]
    body1_geomnum = model.body_geomnum[body1_id]
    body2_geomnum = model.body_geomnum[body2_id]

    # If one of the bodies has no geoms we can skip.
    if body1_geomnum == 0 or body2_geomnum == 0:
        return True

    for geom1_id in range(body1_geomadr, body1_geomadr + body1_geomnum):
        for geom2_id in range(body2_geomadr, body2_geomadr + body2_geomnum):
            if _is_pass_contype_conaffinity_check(model, geom1_id, geom2_id):
                return True

    return False


def _is_pass_contype_conaffinity_check(
    model: mujoco.wrapper.MjModel, geom1_id: int, geom2_id: int
):
    """Returns true if the geoms pass the contype/conaffinity check."""
    return (
        model.geom_contype[geom1_id] & model.geom_conaffinity[geom2_id]
        or model.geom_contype[geom2_id] & model.geom_conaffinity[geom1_id]
    )


def has_collision(
    physics: mjcf.Physics,
    collision_geom_prefix_1: Union[str, Sequence[str]],
    collision_geom_prefix_2: Union[str, Sequence[str]],
    margin: float = _DEFAULT_COLLISION_MARGIN,
) -> bool:
    if isinstance(collision_geom_prefix_1, str):
        collision_geom_prefix_1 = [collision_geom_prefix_1]
    if isinstance(collision_geom_prefix_2, str):
        collision_geom_prefix_2 = [collision_geom_prefix_2]

    for contact in physics.data.contact:
        if contact.dist > margin:
            continue

        geom1_name = physics.model.id2name(contact.geom1, "geom")
        geom2_name = physics.model.id2name(contact.geom2, "geom")

        for pair in itertools.product(collision_geom_prefix_1, collision_geom_prefix_2):
            if (geom1_name.startswith(pair[0]) and geom2_name.startswith(pair[1])) or (
                geom2_name.startswith(pair[0]) and geom1_name.startswith(pair[1])
            ):
                return True

    return False


def has_self_collision(
    physics: mjcf.Physics,
    collision_geom_prefix: Union[str, Sequence[str]],
    margin: float = _DEFAULT_COLLISION_MARGIN,
) -> bool:
    return has_collision(physics, collision_geom_prefix, collision_geom_prefix, margin)



"""Global constants used in the manipulation tasks."""

from typing import Tuple

_Rgba = Tuple[float, float, float, float]

# Predefined RGBA values
RED: _Rgba = (1.0, 0.0, 0.0, 0.3)
GREEN: _Rgba = (0.0, 1.0, 0.0, 0.3)
BLUE: _Rgba = (0.0, 0.0, 1.0, 0.3)
CYAN: _Rgba = (0.0, 1.0, 1.0, 0.3)
MAGENTA: _Rgba = (1.0, 0.0, 1.0, 0.3)
YELLOW: _Rgba = (1.0, 1.0, 0.0, 0.3)

# Invisible group for task-related sites.
TASK_SITE_GROUP: int = 3



"""String constants used to annotate task constructors."""

# Complexity.
EASY = "easy"
HARD = "hard"

# Observation type.
STATE = "features"
VISION = "vision"

# Reward type.
SPARSE = "sparse"
DENSE = "dense"



"""The OpenAI cube."""

from dm_control.entities.props import primitive

from dexterity import _SRC_ROOT

_TEXTURE_PATH = _SRC_ROOT / "manipulation" / "props" / "openai_cube" / "openai_cube.png"


class OpenAICube(primitive.Primitive):
    """A cube with OpenAI letters on each face."""

    def _build(
        self,
        size: float,
        name: str = "openai_cube",
    ) -> None:
        """Builds the cube.

        Args:
            size: The half-length of the cube.
            name: Optional name for the cube prop.
        """
        super()._build(geom_type="box", size=[size] * 3, name=name)

        self.mjcf_model.asset.add(
            "texture",
            name="texture_openai",
            file=str(_TEXTURE_PATH),
            gridsize="3 4",
            gridlayout=".U..LFRB.D..",
        )
        self.mjcf_model.asset.add(
            "material",
            name="material_openai",
            texture="texture_openai",
            specular="1",
            shininess=".3",
            reflectance="0.0",
            rgba="1 1 1 1",
        )

        setattr(self.mjcf_model.find("geom", "geom"), "material", "material_openai")

    @property
    def name(self) -> str:
        return self.mjcf_model.model
